import numpy as np
import numpy.polynomial.legendre as L
import numpy.polynomial.chebyshev as C
from scipy.optimize import least_squares as lsq
import sympy
from sympy import Matrix
from sympy.abc import x

from sympy import integrate, diff
from sympy import legendre
from sympy import lambdify

from library.zoomy_core.model.basemodel import (
    register_sympy_attribute,
    eigenvalue_dict_to_matrix,
)
from library.zoomy_core.model.basemodel import Model
import library.zoomy_core.model.initial_conditions as IC

class CoupledConstrained(Model):
    def __init__(
        self,
        boundary_conditions,
        initial_conditions,
        dimension=2,
        fields=2,
        aux_variables=4,
        parameters={},
        _default_parameters={},
        settings={},
        settings_default={},
    ):
        self.variables = register_sympy_attribute(fields, "q")
        self.n_variables = self.variables.length()
        super().__init__(
            dimension=dimension,
            fields=fields,
            aux_variables=aux_variables,
            parameters=parameters,
            _default_parameters=_default_parameters,
            boundary_conditions=boundary_conditions,
            initial_conditions=initial_conditions,
            settings={**settings_default, **settings},
        )


    def source_implicit(self):
        out = Matrix([0 for i in range(2)])
        u = self.variables[0]
        p = self.variables[1]
        param = self.parameters
        dudt = self.aux_variables.dudx
        dudx = self.aux_variables.dudx
        dpdx = self.aux_variables.dpdx
        f = self.aux_variables.f
        out[0] = dudt + dpdx  + 1
        out[1] = dudx + f
        # out[0] = dudx -1.
        # out[1] = dpdx
        return out



import os
from time import time as gettime

import numpy as np
from attr import define

from typing import Callable
from attrs import define, field

from library.zoomy_core.misc.logger_config import logger



import library.zoomy_core.fvm.flux as flux
import library.zoomy_core.fvm.nonconservative_flux as nonconservative_flux
import library.zoomy_core.misc.io as io
from library.zoomy_core.misc.misc import Zstruct, Settings
import library.zoomy_core.fvm.ode as ode
import library.zoomy_core.fvm.timestepping as timestepping
from library.zoomy_core.transformation.to_numpy import NumpyRuntimeModel


@define(frozen=True, slots=True, kw_only=True)            
class Solver():
    settings: Zstruct = field(factory=lambda: Settings.default())

    def __attrs_post_init__(self):
        defaults = Settings.default()
        defaults.update(self.settings)
        object.__setattr__(self, 'settings', defaults)
        

    def initialize(self, mesh, model):
        # model.boundary_conditions.initialize(
        #     mesh,
        #     model.time,
        #     model.position,
        #     model.distance,
        #     model.variables,
        #     model.aux_variables,
        #     model.parameters,
        #     model.normal,
        # )

        n_variables = model.n_variables
        n_cells = mesh.n_cells
        n_aux_variables = model.aux_variables.length()

        Q = np.empty((n_variables, n_cells), dtype=float)
        Qaux = np.empty((n_aux_variables, n_cells), dtype=float)
        return Q, Qaux
        
    def create_runtime(self, Q, Qaux, mesh, model):      
        mesh.resolve_periodic_bcs(model.boundary_conditions)
        Q, Qaux = np.asarray(Q), np.asarray(Qaux)
        parameters = np.asarray(model.parameter_values)
        runtime_model = NumpyRuntimeModel(model)        
        return Q, Qaux, parameters, mesh, runtime_model

    def get_compute_source(self, mesh, model):
        def compute_source(dt, Q, Qaux, parameters, dQ):
            dQ = model.source(
                    Q[:, :],
                    Qaux[:, :],
                    parameters,
            )
            return dQ

        return compute_source

    def get_compute_source_jacobian_wrt_variables(self, mesh, model):
        def compute_source(dt, Q, Qaux, parameters, dQ):
            dQ = model.source_jacobian_wrt_variables(
                    Q[:, : mesh.n_inner_cells],
                    Qaux[:, : mesh.n_inner_cells],
                    parameters,
                )
            return dQ

        return compute_source
    
    def get_apply_boundary_conditions(self, mesh, model):
        # runtime_bcs = tuple(model.bcs)

        # def apply_boundary_conditions(time, Q, Qaux, parameters):
        #     for i in range(mesh.n_boundary_faces):
        #         i_face = mesh.boundary_face_face_indices[i]
        #         i_bc_func = mesh.boundary_face_function_numbers[i]
        #         q_cell = Q[:, mesh.boundary_face_cells[i]]  # Shape: (Q_dim,)
        #         qaux_cell = Qaux[:, mesh.boundary_face_cells[i]]
        #         normal = mesh.face_normals[:, i_face]
        #         position = mesh.face_centers[i_face, :]
        #         position_ghost = mesh.cell_centers[:, mesh.boundary_face_ghosts[i]]
        #         distance = np.linalg.norm(position - position_ghost)
        #         q_ghost = runtime_bcs[i_bc_func](time, position, distance, q_cell, qaux_cell, parameters, normal)
        #         Q[:,  mesh.boundary_face_ghosts[i]] = q_ghost
        #     return Q
        
        def apply_boundary_conditions(time, Q, Qaux, parameters):
            for i in range(mesh.n_boundary_faces):
                i_face = mesh.boundary_face_face_indices[i]
                i_bc_func = mesh.boundary_face_function_numbers[i]
                q_cell = Q[:, mesh.boundary_face_cells[i]]  # Shape: (Q_dim,)
                qaux_cell = Qaux[:, mesh.boundary_face_cells[i]]
                normal = mesh.face_normals[:, i_face]
                position = mesh.face_centers[i_face, :]
                position_ghost = mesh.cell_centers[:, mesh.boundary_face_ghosts[i]]
                distance = np.linalg.norm(position - position_ghost)
                q_ghost = model.boundary_conditions(i_bc_func, time, position, distance, q_cell, qaux_cell, parameters, normal)
                Q[:,  mesh.boundary_face_ghosts[i]] = q_ghost
            return Q

        return apply_boundary_conditions
    
    def update_q(self, Q, Qaux, mesh, model, parameters):
        """
        Update variables before the solve step.
        """
        # This is a placeholder implementation. Replace with actual logic as needed.
        return Q
    
    def update_qaux(self, Q, Qaux, Qold, Qauxold, mesh, model, parameters, time, dt):
        """
        Update auxiliary variables
        """
        # This is a placeholder implementation. Replace with actual logic as needed.
        return Qaux
    
    def solve(self, mesh, model):
        logger.error(
            "Solver.solve() is not implemented. Please implement this method in the derived class."
        )
        raise NotImplementedError("Solver.solve() must be implemented in derived classes.")

 
@define(frozen=True, slots=True, kw_only=True)            
class HyperbolicSolver(Solver):
    settings: Zstruct = field(factory=lambda: Settings.default())
    compute_dt: Callable = field(factory=lambda: timestepping.adaptive(CFL=0.45))
    num_flux: Callable = field(factory=lambda: flux.Zero())
    nc_flux: Callable = field(factory=lambda: nonconservative_flux.segmentpath())
    time_end: float = 0.1


    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        defaults = Settings.default()
        defaults.output.update(Zstruct(snapshots=10))
        defaults.update(self.settings)
        object.__setattr__(self, 'settings', defaults)
        

    def initialize(self, mesh, model):
        Q, Qaux = super().initialize(mesh, model)
        Q = model.initial_conditions.apply(mesh.cell_centers, Q)
        Qaux = model.aux_initial_conditions.apply(mesh.cell_centers, Qaux)
        return Q, Qaux

    def get_compute_max_abs_eigenvalue(self, mesh, model):
        def compute_max_abs_eigenvalue(Q, Qaux, parameters):
            max_abs_eigenvalue = -np.inf
            i_cellA = mesh.face_cells[0]
            i_cellB = mesh.face_cells[1]
            qA = Q[:, i_cellA]
            qB = Q[:, i_cellB]
            qauxA = Qaux[:, i_cellA]
            qauxB = Qaux[:, i_cellB]
            normal = mesh.face_normals
            evA = model.eigenvalues(qA, qauxA, parameters, normal)
            evB = model.eigenvalues(qB, qauxB, parameters, normal)
            max_abs_eigenvalue = np.maximum(np.abs(evA).max(axis=0), np.abs(evB).max(axis=0))
            return max_abs_eigenvalue
        return compute_max_abs_eigenvalue

    def get_flux_operator(self, mesh, model):
        def flux_operator(dt, Q, Qaux, parameters, dQ):
            compute_num_flux = self.num_flux
            compute_nc_flux = self.nc_flux
            # Initialize dQ as zeros using jax.numpy
            dQ = np.zeros_like(dQ)

            iA = mesh.face_cells[0]
            iB = mesh.face_cells[1]

            qA = Q[:, iA]
            qB = Q[:, iB]
            qauxA = Qaux[:, iA]
            qauxB = Qaux[:, iB]
            normals = mesh.face_normals
            face_volumes = mesh.face_volumes
            cell_volumesA = mesh.cell_volumes[iA]
            cell_volumesB = mesh.cell_volumes[iB]
            svA = mesh.face_subvolumes[:, 0]
            svB = mesh.face_subvolumes[:, 1]

            Dp, Dm, failed = compute_nc_flux(
                qA,
                qB,
                qauxA,
                qauxB,
                parameters,
                normals,
                svA,
                svB,
                face_volumes,
                dt,
                model,
            )
            flux_out = Dm * face_volumes / cell_volumesA
            flux_in = Dp * face_volumes / cell_volumesB

            dQ[:, iA]-= flux_out
            dQ[:, iB]-= flux_in
            return dQ
        return flux_operator

    def solve(self, mesh, model, write_output=True):
        Q, Qaux = self.initialize(mesh, model)
        
        Q, Qaux, parameters, mesh, model = self.create_runtime(Q, Qaux, mesh, model)
        
        # init once with dummy values for dt
        Qaux = self.update_qaux(Q, Qaux, Q, Qaux, mesh, model, parameters, 0.0, 1.0)

        
        if write_output:
            output_hdf5_path = os.path.join(
                self.settings.output.directory, f"{self.settings.output.filename}.h5"
            )
            save_fields = io.get_save_fields(output_hdf5_path, write_all=False)
        else:
            def save_fields(time, time_stamp, i_snapshot, Q, Qaux):
                return i_snapshot

        def run(Q, Qaux, parameters, model):
            iteration = 0.0
            time = 0.0

            i_snapshot = 0.0
            dt_snapshot = self.time_end / (self.settings.output.snapshots - 1)
            if write_output:
                io.init_output_directory(
                    self.settings.output.directory, self.settings.output.clean_directory
                )
                mesh.write_to_hdf5(output_hdf5_path)
                io.save_settings(self.settings)
            i_snapshot = save_fields(time, 0.0, i_snapshot, Q, Qaux)

            Qnew = Q
            Qauxnew = Qaux

            compute_max_abs_eigenvalue = self.get_compute_max_abs_eigenvalue(mesh, model)
            flux_operator = self.get_flux_operator(mesh, model)
            source_operator = self.get_compute_source(mesh, model)
            boundary_operator = self.get_apply_boundary_conditions(mesh, model)
            Qnew = boundary_operator(time, Qnew, Qaux, parameters)


            cell_inradius_face = np.minimum(mesh.cell_inradius[mesh.face_cells[0, :]], mesh.cell_inradius[mesh.face_cells[1,:]])

            while time < self.time_end:
                Q = Qnew
                Qaux = Qauxnew
                
                dt = self.compute_dt(
                    Q, Qaux, parameters, cell_inradius_face, compute_max_abs_eigenvalue
                )

                Q1 = ode.RK1(flux_operator, Q, Qaux, parameters, dt)
                Q2 = ode.RK1(
                    source_operator,
                    Q1,
                    Qaux,
                    parameters,
                    dt,
                )

                Q3 = boundary_operator(time, Q2, Qaux, parameters)
                
                # Update solution and time
                time += dt
                iteration += 1

                time_stamp = (i_snapshot) * dt_snapshot
                
                Qnew = self.update_q(Q3, Qaux, mesh, model, parameters)
                Qauxnew = self.update_qaux(Qnew, Qaux, Q, Qaux, mesh, model, parameters, time, dt)


                i_snapshot = save_fields(time, time_stamp, i_snapshot, Qnew, Qauxnew)

                if iteration % 10 == 0:
                    logger.info(
                        f"iteration: {int(iteration)}, time: {float(time):.6f}, "
                        f"dt: {float(dt):.6f}, next write at time: {float(time_stamp):.6f}"
                    )

            return Qnew, Qaux

        time_start = gettime()
        Qnew, Qaux = run(Q, Qaux, parameters, model)
        time = gettime() - time_start
        logger.info(f"Finished simulation with in {time:.3f} seconds")
        return Qnew, Qaux
    



import numpy as np
import os
import logging

import sympy
from sympy import Symbol, Matrix, lambdify, transpose, Abs, sqrt

from sympy import zeros, ones

from attr import define, field
from typing import Optional
from types import SimpleNamespace

from library.zoomy_core.model.boundary_conditions import BoundaryConditions, Extrapolation
from library.zoomy_core.model.initial_conditions import InitialConditions, Constant
from library.zoomy_core.misc.custom_types import FArray
from library.zoomy_core.model.basemodel import Model, eigenvalue_dict_to_matrix

from library.zoomy_core.misc.misc import Zstruct

@define(frozen=True, slots=True, kw_only=True)
class ShallowWaterEquationsWithTopo1D(Model):
    dimension: int = 1
    variables: Zstruct = field(default=3)
    aux_variables: Zstruct = field(default=0)

@define(frozen=True, slots=True, kw_only=True)
class ShallowWaterEquationsWithTopo(Model):
    dimension: int=2
    variables: Zstruct = field(default=4)
    aux_variables: Zstruct = field(default=0)
    _default_parameters: dict = field(
        init=False,
        factory=lambda: {"g": 9.81, "ex": 0.0, "ey": 0.0, "ez": 1.0}
        )
    
    def __attrs_post_init__(self):
        if type(self.variables)==int:
            object.__setattr__(self, "variables",self.dimension+2)
        super().__attrs_post_init__()
        
    def compute_hinv(self):
        h = self.variables[1]
        return 1 / h
        
    def get_primitives(self):
        dim = self.dimension
        b = self.variables[0]
        h = self.variables[1]
        hinv = self.compute_hinv()
        U = [self.variables[2 + i] * hinv for i in range(dim)]
        return b, h, U, hinv


    def project_2d_to_3d(self):
        out = Matrix([0 for i in range(5)])
        dim = self.dimension
        x = self.position[0]
        y = self.position[1]
        z = self.position[2]
        b, h, U, hinv = self.get_primitives()
        rho_w = 1000.
        g = 9.81
        out[0] = h
        out[1] = U[0]
        out[2] = 0 if dim == 1 else U[1]
        out[3] = 0
        out[4] = rho_w * g * h * (1-z)
        return out

        

    def flux(self):
        dim = self.dimension
        h = self.variables[1]
        U = Matrix([hu / h for hu in self.variables[2:2+dim]])
        g = self.parameters.g
        I = Matrix.eye(dim)
        F = Matrix.zeros(self.variables.length(), dim)
        F[1, :] = (h * U).T
        F[2:, :] = h * U * U.T + g/2 * h**2 * I
        return [F[:, d] for d in range(dim)]
    
    def nonconservative_matrix(self):
        nc_x = Matrix([[0 for i in range(self.n_variables)] for j in range(self.n_variables)])
        nc_y = Matrix([[0 for i in range(self.n_variables)] for j in range(self.n_variables)])
        b, h, U, hinv = self.get_primitives()
        p = self.parameters
        nc_x[1, 0] = p.g * h 
        nc_y[1, 0] = p.g * h
        return [nc_x, nc_y][:self.dimension]

    
    def source(self):
        out = Matrix([0 for i in range(self.n_variables)])
        return out
    
    def chezy(self):
        assert "C" in vars(self.parameters)
        dim = self.dimension
        out = Matrix([0 for i in range(self.n_variables)])
        h = self.variables[0]
        hU = self.variables[1:1+dim]
        U = Matrix([hu / h for hu in hU])
        p = self.parameters
        u_sq = sqrt(U.dot(U))
        out[2:2+dim] = -1.0 / p.C**2 * U * u_sq
        return out
    
        return self._simplify(eigenvalue_dict_to_matrix(A.eigenvals()))
    


import numpy as np
import numpy.polynomial.legendre as L
import numpy.polynomial.chebyshev as C
from scipy.optimize import least_squares as lsq
import sympy
from sympy import Matrix, Piecewise, sqrt
from sympy.abc import x

from sympy import integrate, diff
from sympy import legendre
from sympy import lambdify

from attrs import define, field
from typing import Union



from library.zoomy_core.model.basemodel import (
    register_sympy_attribute,
    eigenvalue_dict_to_matrix,
)
from library.zoomy_core.model.basemodel import Model
import library.zoomy_core.model.initial_conditions as IC
from library.zoomy_core.model.models.basismatrices import Basismatrices
from library.zoomy_core.model.models.basisfunctions import Legendre_shifted, Basisfunction



@define(frozen=True, slots=True, kw_only=True)
class ShallowMoments2d(Model):
    dimension: int = 2
    level: int
    variables: Union[list, int] = field(init=False)
    aux_variables: Union[list, int] = field(default=2)
    basisfunctions: Union[Basisfunction, type[Basisfunction]] = field(default=Legendre_shifted)
    basismatrices: Basismatrices = field(init=False)

    _default_parameters: dict = field(
        init=False,
        factory=lambda: {"g": 9.81, "ex": 0.0, "ey": 0.0, "ez": 1.0}
    )

    def __attrs_post_init__(self):
        object.__setattr__(self, "variables", ((self.level+1)*self.dimension)+1)
        super().__attrs_post_init__()
        aux_variables = self.aux_variables
        aux_var_list = aux_variables.keys()
        if not aux_variables.contains("dudx"):
            aux_var_list += ["dudx"]
        if self.dimension == 2 and not aux_variables.contains("dvdy"):
            aux_var_list += ["dvdy"]
        object.__setattr__(self, "aux_variables", register_sympy_attribute(aux_var_list, "qaux_"))

        # Recompute basis matrices
        object.__setattr__(self, "basisfunctions", self.basisfunctions(level=self.level))
        basismatrices = Basismatrices(self.basisfunctions)
        basismatrices.compute_matrices(self.level)
        object.__setattr__(self, "basismatrices", basismatrices)
        



    def project_2d_to_3d(self):
        out = Matrix([0 for i in range(6)])
        level = self.level
        offset = level+1
        x = self.position[0]
        y = self.position[1]
        z = self.position[2]
        h = self.variables[0]
        a = [self.variables[1+i]/h for i in range(offset)]
        dhdx = self.aux_variables[0]
        dadx = [self.aux_variables[1+i] for i in range(offset)]
        
        psi = [self.basisfunctions.eval_psi(k, z) for k in range(level+1)]
        phi = [self.basisfunctions.eval(k, z) for k in range(level+1)]

        rho_w = 1000.
        g = 9.81
        u_3d = self.basismatrices.basisfunctions.reconstruct_velocity_profile_at(a, z)
        v_3d = 0
        b = 0
        dbdx = 0
        def dot(a, b):
            s = 0
            for i in range(len(a)):
                s += a[i] * b[i]
            return s
        w_3d = - dhdx * dot(a,psi) - h * dot(dadx,psi) + dot(a, phi) * (z * dhdx + dbdx)
        if self.dimension == 2:
            beta = [self.variables[1+offset+i]/h for i in range(offset)]
            v_3d = self.basismatrices.basisfunctions.reconstruct_velocity_profile_at(beta, z)

        b = 0
        out[0] = b
        out[1] = h
        out[2] = u_3d
        out[3] = v_3d
        out[4] = w_3d
        out[5] = rho_w * g * h * (1-z)

        return out

    def flux(self):
        offset = self.level + 1
        flux_x = Matrix([0 for i in range(self.n_variables)])
        flux_y = Matrix([0 for i in range(self.n_variables)])
        h = self.variables[0]
        ha = self.variables[1 : 1 + self.level + 1]
        p = self.parameters
        flux_x[0] = ha[0]
        flux_x[1] = p.g * p.ez * h * h / 2
        for k in range(self.level + 1):
            for i in range(self.level + 1):
                for j in range(self.level + 1):
                    # TODO avoid devision by zero
                    flux_x[k + 1] += (
                        ha[i]
                        * ha[j]
                        / h
                        * self.basismatrices.A[k, i, j] / self.basismatrices.M[k, k]
                    )
        if self.dimension == 2:
            hb = self.variables[1 + self.level + 1 : 1 + 2 * (self.level + 1)]

            for k in range(self.level + 1):
                for i in range(self.level + 1):
                    for j in range(self.level + 1):
                        # TODO avoid devision by zero
                        flux_x[k + 1 + offset] += (
                            hb[i]
                            * ha[j]
                            / h
                            * self.basismatrices.A[k, i, j]/ self.basismatrices.M[k, k]
                        )

            flux_y[0] = hb[0]
            flux_y[1 + offset] = p.g * p.ez * h * h / 2
            for k in range(self.level + 1):
                for i in range(self.level + 1):
                    for j in range(self.level + 1):
                        # TODO avoid devision by zero
                        flux_y[k + 1] += (
                            hb[i]
                            * ha[j]
                            / h
                            * self.basismatrices.A[k, i, j]/ self.basismatrices.M[k, k]
                        )
            for k in range(self.level + 1):
                for i in range(self.level + 1):
                    for j in range(self.level + 1):
                        # TODO avoid devision by zero
                        flux_y[k + 1 + offset] += (
                            hb[i]
                            * hb[j]
                            / h
                            * self.basismatrices.A[k, i, j]/ self.basismatrices.M[k, k]
                    )
        return [flux_x, flux_y]

    def nonconservative_matrix(self):
        offset = self.level + 1
        nc_x = Matrix([[0 for i in range(self.n_variables)] for j in range(self.n_variables)])
        nc_y = Matrix([[0 for i in range(self.n_variables)] for j in range(self.n_variables)])
        h = self.variables[0]
        ha = self.variables[1 : 1 + self.level + 1]
        p = self.parameters
        um = ha[0] / h

        for k in range(1, self.level + 1):
            nc_x[k + 1, k + 1] += um
        for k in range(self.level + 1):
            for i in range(1, self.level + 1):
                for j in range(1, self.level + 1):
                    nc_x[k + 1, i + 1] -= (
                        ha[j]
                        / h
                        * self.basismatrices.B[k, i, j]/ self.basismatrices.M[k, k]
                    )

                        
        if self.dimension ==  2:
            hb = self.variables[1 + offset : 1 + offset + self.level + 1]
            vm = hb[0] / h
            for k in range(1, self.level + 1):
                nc_y[k + 1, k + 1 + offset] += um
            for k in range(self.level + 1):
                for i in range(1, self.level + 1):
                    for j in range(1, self.level + 1):
                        nc_y[k + 1, i + 1 + offset] -= (
                            ha[j]
                            / h
                            * self.basismatrices.B[k, i, j]/ self.basismatrices.M[k, k]
                        )

            for k in range(1, self.level + 1):
                nc_x[k + 1 + offset, k + 1] += vm
                nc_y[k + 1 + offset, k + 1 + offset] += vm
            for k in range(self.level + 1):
                for i in range(1, self.level + 1):
                    for j in range(1, self.level + 1):
                        nc_x[k + 1 + offset, i + 1] -= (
                            hb[j]
                            / h
                            * self.basismatrices.B[k, i, j]
                        )
                        nc_y[k + 1 + offset, i + 1 + offset] -= (
                            hb[j]
                            / h
                            * self.basismatrices.B[k, i, j]/ self.basismatrices.M[k, k]
                        )
        return [-nc_x, -nc_y]

    def eigenvalues(self):
        # we delete heigher order moments (level >= 2) for analytical eigenvalues
        offset = self.level + 1
        A = self.normal[0] * self.quasilinear_matrix()[0]
        for d in range(1, self.dimension):
            A += self.normal[d] * self.quasilinear_matrix()[d]
        alpha_erase = self.variables[2 : 2 + self.level]
        beta_erase = self.variables[2 + offset : 2 + offset + self.level]
        for alpha_i in alpha_erase:
            A = A.subs(alpha_i, 0)
        for beta_i in beta_erase:
            A = A.subs(beta_i, 0)
        return eigenvalue_dict_to_matrix(A.eigenvals())


    def source(self):
        out = Matrix([0 for i in range(self.n_variables)])
        return out
    
    def gravity(self):
        out = Matrix([0 for i in range(self.n_variables)])
        out[1] = -self.parameters.g * self.parameters.ex * self.variables[0]
        if self.dimension == 2:
            offset = self.level + 1
            out[1 + offset] = -self.parameters.g * self.parameters.ey * self.variables[0]
        return out
    
    def newtonian_turbulent(self):
        p = self.parameters
        nut1 = [
            1.06245397e-05,
            -8.64966128e-06,
            -4.24655215e-06,
            1.51861028e-06,
            2.25140517e-06,
            1.81867029e-06,
            -1.02154323e-06,
            -1.78795289e-06,
            -5.07515843e-07,
        ]
        nut2 = np.array(
            [
                0.21923893,
                -0.04171894,
                -0.05129916,
                -0.04913612,
                -0.03863209,
                -0.02533469,
                -0.0144186,
                -0.00746847,
                -0.0031811,
                -0.00067986,
                0.0021782,
            ]
        )
        nut2 = nut2 / nut2[0] * 1.06245397 * 10 ** (-5)
        nut3 = [
            1.45934315e-05,
            -1.91969629e-05,
            5.80456268e-06,
            -5.13207491e-07,
            2.29489571e-06,
            -1.24361978e-06,
            -2.78720732e-06,
            -2.01469118e-07,
            1.24957663e-06,
        ]
        nut4 = [
            1.45934315e-05,
            -1.45934315e-05 * 3 / 4,
            -1.45934315e-05 * 1 / 4,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
        ]
        nut5 = [p.nut, -p.nut * 3 / 4, -p.nut * 1 / 4, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        nut = nut5
        out = Matrix([0 for i in range(self.n_variables)])
        h = self.variables[0]
        ha = self.variables[1 : 1 + self.level + 1]
        p = self.parameters
        for k in range(1 + self.level):
            for i in range(1 + self.level):
                for j in range(1 + self.level):
                    out[1 + k] += (
                        -p.c_nut
                        * nut[j]
                        / h
                        * ha[i]
                        / h
                        * self.basismatrices.DT[k, i, j]/ self.basismatrices.M[k, k]
                    )
        return  out
    
    def newtonian_boundary_layer_classic(self):
        assert "nu" in vars(self.parameters)
        assert "eta" in vars(self.parameters)
        out = Matrix([0 for i in range(self.n_variables)])
        h = self.variables[0]
        ha = self.variables[1 : 1 + self.level + 1]
        p = self.parameters
        phi_0 = [
            self.basismatrices.basisfunctions.eval(i, 0.0)
            for i in range(self.level + 1)
        ]
        dphidx_0 = [
            (diff(self.basismatrices.basisfunctions.eval(i, x), x)).subs(x, 0.0)
            for i in range(self.level + 1)
        ]
        tau_bot = 0
        for i in range(1 + self.level):
            tau_bot += ha[i] / h * dphidx_0[i]
        for k in range(1 + self.level):
            out[k + 1] = (
                -p.c_bl
                * p.eta
                * (p.nu + p.nut_bl)
                / h
                * tau_bot
                * phi_0[k]/ self.basismatrices.M[k, k]
            )
        return  out

    def newtonian(self):
        assert "nu" in vars(self.parameters)
        out = Matrix([0 for i in range(self.n_variables)])
        offset = self.level + 1
        h = self.variables[0]
        ha = self.variables[1 : 1 + self.level + 1]
        p = self.parameters
        for k in range(1 + self.level):
            for i in range(1 + self.level):
                out[1 + k] += (
                    -p.nu
                    / h
                    * ha[i]
                    / h
                    * self.basismatrices.D[i, k]/ self.basismatrices.M[k, k]
                )
        if self.dimension == 2:
            hb = self.variables[1 + offset : 1 + self.level + 1 + offset]
            for k in range(1 + self.level):
                for i in range(1 + self.level):
                    out[1 + k + offset] += (
                        -p.nu
                        / h
                        * hb[i]
                        / h
                        * self.basismatrices.D[i, k]/ self.basismatrices.M[k, k]
                    )

        return out
    
    def newtonian_turbulent_algebraic(self):
        assert "nu" in vars(self.parameters)
        assert "l_bl" in vars(self.parameters)
        assert "l_turb" in vars(self.parameters)
        assert "kappa" in vars(self.parameters)
        out = Matrix([0 for i in range(self.n_variables)])
        offset = self.level + 1
        h = self.variables[0]
        a = [_ha / h for _ha in self.variables[1 : 1 + self.level + 1]]
        p = self.parameters
        dU_dx = a[0] / (p.l_turb * h)
        abs_dU_dx = sympy.Piecewise((dU_dx, dU_dx >=0), (-dU_dx, True))
        for k in range(1 + self.level):
            out[1 + k] += (
                -(p.nu + p.kappa * sympy.sqrt(p.nu * abs_dU_dx) * p.l_bl * ( 1-p.l_bl)) * dU_dx * self.basismatrices.phib[k] / h/ self.basismatrices.M[k, k]
            )
            for i in range(1 + self.level):
                out[1 + k] += (
                    -p.nu
                    / h
                    * a[i]
                    * self.basismatrices.D[i, k]/ self.basismatrices.M[k, k]
                )
                out[1 + k] += (
                    -p.kappa * sympy.sqrt(p.nu * abs_dU_dx)
                    / h
                    * a[i]
                    * (self.basismatrices.Dxi[i, k] - self.basismatrices.Dxi2[i, k])/ self.basismatrices.M[k, k]
                )
        if self.dimension == 2:
            b = [_hb / h for _hb in self.variables[1 + offset : 1 + self.level + 1 + offset] ]
            dV_dy = b[0] / (p.l_turb * h)
            abs_dV_dy = sympy.Piecewise((dV_dy, dV_dy >=0), (-dV_dy, True))
            for k in range(1 + self.level):
                out[1 + k + offset] += (
                    -(p.nu + p.kappa * sympy.sqrt(p.nu * abs_dV_dy) * p.l_bl * ( 1-p.l_bl)) * dV_dy * self.basismatrices.phib[k] / h/ self.basismatrices.M[k, k]
                )
                for i in range(1 + self.level):
                    out[1 + k + offset] += (
                        -p.nu
                        / h
                        * b[i]
                        * self.basismatrices.D[i, k]/ self.basismatrices.M[k, k]
                    )
                    out[1 + k + offset] += (
                    -p.kappa * sympy.sqrt(p.nu * abs_dV_dy)
                    / h
                    * b[i]
                    * (self.basismatrices.Dxi[i, k] - self.basismatrices.Dxi2[i, k])/ self.basismatrices.M[k, k]
                )

        return out
    
    def regress_against_power_profile(self):
        """
        :gui:
            - requires_parameter: ('lamda', 0.0)
            - requires_parameter: ('rho', 1.0)
        """
        assert "r_pp" in vars(self.parameters)
        out = Matrix([0 for i in range(self.n_variables)])
        offset = self.level+1
        h = self.variables[0]
        ha = self.variables[1 : 1 + self.level + 1]
        p = self.parameters
        ub = 0
        Z = np.linspace(0,1,100)
        U = 1-(1-Z)**8
        power_profile_coefs = self.basisfunctions.project_onto_basis(U)
        for i in range(1 + self.level):
            ub += ha[i] / h
        for k in range(1, 1 + self.level):
            out[1 + k] += (
                -p.r_pp * sympy.Piecewise((ub, ub >=0), (-ub, True)) * (ha[i] - ha[0]*power_profile_coefs[i])/ self.basismatrices.M[k, k]
            )
        return out

    def slip_mod(self):
        """
        :gui:
            - requires_parameter: ('lamda', 0.0)
            - requires_parameter: ('rho', 1.0)
        """
        assert "lamda" in vars(self.parameters)
        assert "rho" in vars(self.parameters)
        out = Matrix([0 for i in range(self.n_variables)])
        offset = self.level+1
        h = self.variables[0]
        ha = self.variables[1 : 1 + self.level + 1]
        p = self.parameters
        ub = 0
        for i in range(1 + self.level):
            ub += ha[i] / h
        for k in range(1, 1 + self.level):
            out[1 + k] += (
                -p.c_slipmod / p.lamda / p.rho * ub / self.basismatrices.M[k, k]
            )
        if self.dimension == 2:
            hb = self.variables[1+offset : 1+offset + self.level + 1]
            vb = 0
            for i in range(1 + self.level):
                vb += hb[i] / h
            for k in range(1, 1 + self.level):
                out[1+offset+k] += (
                    -1.0 * p.c_slipmod / p.lamda / p.rho * vb/ self.basismatrices.M[k, k]
                )
        return out

    def newtonian_boundary_layer(self):
        assert "nu" in vars(self.parameters)
        out = Matrix([0 for i in range(self.n_variables)])
        offset = self.level + 1
        h = self.variables[0]
        ha = self.variables[1 : 1 + self.level + 1]
        p = self.parameters
        phi_0 = [self.basismatrices.eval(i, 0.0) for i in range(self.level + 1)]
        dphidx_0 = [
            (diff(self.basismatrices.eval(i, x), x)).subs(x, 0.0)
            for i in range(self.level + 1)
        ]
        for k in range(1 + self.level):
            for i in range(1 + self.level):
                out[1 + k] += (
                    -p.nu
                    / h
                    * ha[i]
                    / h
                    * phi_0[k]
                    * dphidx_0[i]/ self.basismatrices.M[k, k]
                )
        if self.dimension==2:
            hb = self.variables[1 + offset : 1 + self.level + 1 + offset]
            for k in range(1 + self.level):
                for i in range(1 + self.level):
                    out[1 + k + offset] += (
                        -p.nu
                        / h
                        * hb[i]
                        / h
                        * phi_0[k]
                        * dphidx_0[i]/ self.basismatrices.M[k, k]
                    )
        return out

    def sindy(self):
        assert "nu" in vars(self.parameters)
        out = Matrix([0 for i in range(self.n_variables)])
        offset = self.level + 1
        h = self.variables[0]
        ha = self.variables[1 : 1 + self.level + 1]
        hb = self.variables[1 + offset : 1 + self.level + 1 + offset]
        p = self.parameters
        out[1] += (
            p.C1 * sympy.Abs(ha[0] / h)
            + p.C2 * sympy.Abs(ha[1] / h)
            + p.C3 * sympy.Abs(ha[0] / h) ** (7 / 3)
            + p.C4 * sympy.Abs(ha[1] / h) ** (7 / 3)
        )
        out[2] += (
            p.C5 * sympy.Abs(ha[0] / h)
            + p.C6 * sympy.Abs(ha[1] / h)
            + p.C7 * sympy.Abs(ha[0] / h) ** (7 / 3)
            + p.C8 * sympy.Abs(ha[1] / h) ** (7 / 3)
        )
        out[3] += (
            p.C1 * sympy.Abs(ha[0] / h)
            + p.C2 * sympy.Abs(ha[1] / h)
            + p.C3 * sympy.Abs(ha[0] / h) ** (7 / 3)
            + p.C4 * sympy.Abs(ha[1] / h) ** (7 / 3)
        )
        out[4] += (
            p.C5 * sympy.Abs(ha[0] / h)
            + p.C6 * sympy.Abs(ha[1] / h)
            + p.C7 * sympy.Abs(ha[0] / h) ** (7 / 3)
            + p.C8 * sympy.Abs(ha[1] / h) ** (7 / 3)
        )
        return out

    def slip(self):
        assert "lamda" in vars(self.parameters)
        assert "rho" in vars(self.parameters)
        out = Matrix([0 for i in range(self.n_variables)])
        offset = self.level + 1
        h = self.variables[0]
        h = self.variables[0]
        ha = self.variables[1 : 1 + self.level + 1]
        p = self.parameters
        for k in range(1 + self.level):
            for i in range(1 + self.level):
                out[1 + k] += (
                    -1.0 / p.lamda / p.rho * ha[i] / h / self.basismatrices.M[k, k]
                )

        if self.dimension == 2:
            hb = self.variables[1 + offset : 1 + self.level + 1 + offset]
            for k in range(1 + self.level):
                for i in range(1 + self.level):
                    out[1 + k + offset] += (
                        -1.0 / p.lamda / p.rho/ self.basismatrices.M[k, k]
                    )
        return out

    def chezy(self):
        assert "C" in vars(self.parameters)
        out = Matrix([0 for i in range(self.n_variables)])
        h = self.variables[0]
        ha = self.variables[1 : 1 + self.level + 1]
        p = self.parameters
        tmp = 0
        if self.dimension == 1:
            for i in range(1 + self.level):
                for j in range(1 + self.level):
                    tmp += ha[i] * ha[j] / h / h
            sqrt = sympy.sqrt(tmp)
            for k in range(1 + self.level):
                for l in range(1 + self.level):
                    out[1 + k] += (
                        -1.0 / (p.C**2) * ha[l] * sqrt / h/ self.basismatrices.M[k, k]
                    )
        if self.dimension == 2:
            offset = self.level + 1
            hb = self.variables[1 + offset : 1 + self.level + 1 + offset]
            for i in range(1 + self.level):
                for j in range(1 + self.level):
                    tmp += ha[i] * ha[j] / h / h + hb[i] * hb[j] / h / h
            sqrt = sympy.sqrt(tmp)
            for k in range(1 + self.level):
                for l in range(1 + self.level):
                    out[1 + k] += (
                        -1.0 / (p.C**2) * ha[l] * sqrt / h/ self.basismatrices.M[k, k]
                    )
                    out[1 + k + offset] += (
                        -1.0 / (p.C**2) * hb[l] * sqrt / h/ self.basismatrices.M[k, k]
                    )

        return out


def reconstruct_uvw(Q, grad, lvl, phi, psi):
    """
    returns functions u(z), v(z), w(z)
    """
    offset = lvl + 1
    h = Q[0]
    alpha = Q[1 : 1 + offset] / h
    beta = Q[1 + offset : 1 + 2 * offset] / h
    dhalpha_dx = grad[1 : 1 + offset, 0]
    dhbeta_dy = grad[1 + offset : 1 + 2 * offset, 1]

    def u(z):
        u_z = 0
        for i in range(lvl + 1):
            u_z += alpha[i] * phi(z)[i]
        return u_z

    def v(z):
        v_z = 0
        for i in range(lvl + 1):
            v_z += beta[i] * phi(z)[i]
        return v_z

    def w(z):
        basis_0 = psi(0)
        basis_z = psi(z)
        u_z = 0
        v_z = 0
        grad_h = grad[0, :]
        # grad_hb = grad[-1, :]
        grad_hb = np.zeros(grad[0, :].shape)
        result = 0
        for i in range(lvl + 1):
            u_z += alpha[i] * basis_z[i]
            v_z += beta[i] * basis_z[i]
        for i in range(lvl + 1):
            result -= dhalpha_dx[i] * (basis_z[i] - basis_0[i])
            result -= dhbeta_dy[i] * (basis_z[i] - basis_0[i])

        result += u_z * (z * grad_h[0] + grad_hb[0])
        result += v_z * (z * grad_h[1] + grad_hb[1])
        return result

    return u, v, w


def generate_velocity_profiles(
    Q,
    centers,
    model: Model,
    list_of_positions: list[np.ndarray],
):
    def find_closest_element(centers, pos):
        assert centers.shape[1] == np.array(pos).shape[0]
        return np.argmin(np.linalg.norm(centers - pos, axis=1))

    # find the closest element to the given position
    vertices = []
    for pos in list_of_positions:
        vertex = find_closest_element(centers, pos)
        vertices.append(vertex)

    Z = np.linspace(0, 1, 100)
    list_profiles = []
    list_means = []
    level = int((model.n_variables - 1) / model.dimension) - 1
    offset = level + 1
    list_h = []
    for vertex in vertices:
        profiles = []
        means = []
        for d in range(model.dimension):
            q = Q[vertex, :]
            h = q[0]
            coefs = q[1 + d * offset : 1 + (d + 1) * offset] / h
            profile = model.basis.basis.reconstruct_velocity_profile(coefs, Z=Z)
            mean = coefs[0]
            profiles.append(profile)
            means.append(mean)
        list_profiles.append(profiles)
        list_means.append(means)
        list_h.append(h)
    return list_profiles, list_means, list_of_positions, Z, list_h


if __name__ == "__main__":
    # basis = Legendre_shifted(1)
    # basis = Spline()
    # basis = OrthogonalSplineWithConstant(degree=2, knots=[0, 0.1, 0.3,0.5, 1,1])
    # basis=OrthogonalSplineWithConstant(degree=1, knots=[0,0, 0.02, 0.04, 0.06, 0.08, 0.1,  1])
    # basis=OrthogonalSplineWithConstant(degree=1, knots=[0,0, 0.1, 1])
    # basis.plot()

    # basis = Legendre_shifted(basis=Legendre_shifted(order=8))
    # f = basis.enforce_boundary_conditions()
    # q = np.array([[1., 0.1, 0., 0., 0., 0.], [1., 0.1, 0., 0., 3., 0.]])
    # print(f(q))

    # basis =Legendre_shifted(order=8)
    # basis.plot()
    # z = np.linspace(0,1,100)
    # f = basis.get_lambda(1)
    # print(f(z), f(1.0))
    # f = basis.get_lambda(1)
    # print(f(z))

    # X = np.linspace(0,1,100)
    # coef = np.array([0.2, -0.01, -0.1, -0.05, -0.04])
    # U = basis.basis.reconstruct_velocity_profile(coef, Z=X)
    # coef2 = coef*2
    # factor = 1.0 / 0.2
    # coef3  = coef * factor
    # U2 = basis.basis.reconstruct_velocity_profile(coef2, Z=X)
    # U3 = basis.basis.reconstruct_velocity_profile(coef3, Z=X)
    # fig, ax = plt.subplots()
    # ax.plot(U, X)
    # ax.plot(U2, X)
    # ax.plot(U3, X)
    # plt.show()

    # X = np.linspace(0,1,100)
    # nut = 10**(-5)
    # coef = np.array([nut, -nut, 0, 0, 0, 0, 0 ])
    # U = basis.basis.reconstruct_velocity_profile(coef, Z=X)
    # fig, ax = plt.subplots()
    # ax.plot(U, X)
    # plt.show()

    # nut = np.load('/home/ingo/Git/sms/nut_nut2.npy')
    # y = np.load('/home/ingo/Git/sms/nut_y2.npy')
    # coef = basis.basis.reconstruct_alpha(nut, y)
    # coef_offset = np.sum(coef)
    # coef[0] -= coef_offset
    # print(coef)
    # X = np.linspace(0,1,100)
    # _nut = basis.basis.reconstruct_velocity_profile(coef, Z=X)
    # fig, ax = plt.subplots()
    # ax.plot(_nut, X)
    # plt.show()

    basis = Legendre_shifted(basis=Legendre_shifted(level=2))
    basis.compute_matrices(2)
    print(basis.D)

@define(frozen=True, slots=True, kw_only=True)
class ShallowMoments(ShallowMoments2d):
    dimension: int = 1



import numpy as np
import os
# import logging

import sympy
from sympy import Symbol, Matrix, lambdify, transpose, Abs, sqrt

from sympy import zeros, ones

from attr import define
from typing import Optional
from types import SimpleNamespace

from library.zoomy_core.model.boundary_conditions import BoundaryConditions, Extrapolation
from library.zoomy_core.model.initial_conditions import InitialConditions, Constant
from library.zoomy_core.misc.custom_types import FArray

# from library.zoomy_core.misc import vectorize  # type: ignore
from library.zoomy_core.model.basemodel import Model


class Advection(Model):
    def flux(self):
        if self.dimension == 1:
            F = Matrix([0 for i in range(self.n_variables)])
            for i_field in range(self.n_variables):
                F[i_field] = self.variables[i_field] * self.parameters[0]
            return [F]
        elif self.dimension == 2:
            F = Matrix([0 for i in range(self.n_variables)])
            G = Matrix([0 for i in range(self.n_variables)])
            for i_field in range(self.n_variables):
                F[i_field] = self.variables[i_field] * self.parameters[0]
                G[i_field] = self.variables[i_field] * self.parameters[1]
            return [F, G]
        elif self.dimension == 3:
            F = Matrix([0 for i in range(self.n_variables)])
            G = Matrix([0 for i in range(self.n_variables)])
            H = Matrix([0 for i in range(self.n_variables)])
            for i_field in range(self.n_variables):
                F[i_field] = self.variables[i_field] * self.parameters[0]
                G[i_field] = self.variables[i_field] * self.parameters[1]
                H[i_field] = self.variables[i_field] * self.parameters[2]
            return [F, G, H]
        else:
            assert False

    # def eigenvalues(self):
    #     assert self.normal.shape[0] == self.parameters.shape[0]
    #     ev = self.normal[0] * self.parameters[0]
    #     for d in range(1, self.dimension):
    #         ev += self.normal[d] * self.parameters[d]
    #     self.sympy_eigenvalues = Matrix[[ev for i in range(self.n_variables)]]



import os
import numpy as np
import sympy
from sympy import integrate, diff, Matrix
from sympy.abc import z
from time import time as get_time

from scipy.optimize import least_squares as lsq

from library.zoomy_core.model.models.basisfunctions import Legendre_shifted


class Basismatrices:
    def __init__(self, basis=Legendre_shifted(), use_cache=True, cache_path=".cache"):
        self.basisfunctions = basis
        self.use_cache = use_cache
        self.cache_dir = cache_path
        self.cache_subdir = f"basismatrices/{basis.name}/{basis.level}"

    def load_cached_matrices(self):
        main_dir = os.getenv("ZOOMY_DIR")
        path = os.path.join(os.path.join(main_dir, self.cache_dir), self.cache_subdir)
        failed = False
        try:
            self.phib = np.load(os.path.join(path, "phib.npy"))
            self.M = np.load(os.path.join(path, "M.npy"))
            self.A = np.load(os.path.join(path, "A.npy"))
            self.B = np.load(os.path.join(path, "B.npy"))
            self.D = np.load(os.path.join(path, "D.npy"))
            self.Dxi = np.load(os.path.join(path, "Dxi.npy"))
            self.Dxi2 = np.load(os.path.join(path, "Dxi2.npy"))
            self.DD = np.load(os.path.join(path, "DD.npy"))
            self.D1 = np.load(os.path.join(path, "D1.npy"))
            self.DT = np.load(os.path.join(path, "DT.npy"))
        except:
            failed = True
        return failed

    def save_cached_matrices(self):
        main_dir = os.getenv("ZOOMY_DIR")
        path = os.path.join(os.path.join(main_dir, self.cache_dir), self.cache_subdir)
        os.makedirs(path, exist_ok=True)
        np.save(os.path.join(path, "phib"), self.phib)
        np.save(os.path.join(path, "M"), self.M)
        np.save(os.path.join(path, "A"), self.A)
        np.save(os.path.join(path, "B"), self.B)
        np.save(os.path.join(path, "D"), self.D)
        np.save(os.path.join(path, "Dxi"), self.Dxi)
        np.save(os.path.join(path, "Dxi2"), self.Dxi2)
        np.save(os.path.join(path, "DD"), self.DD)
        np.save(os.path.join(path, "D1"), self.D1)
        np.save(os.path.join(path, "DT"), self.DT)
        


    def _compute_matrices(self, level):
        start = get_time()
        # object is key here, as we need to have a symbolic representation of the fractions.
        self.phib = np.empty((level + 1), dtype=object)
        self.M = np.empty((level + 1, level + 1), dtype=object)
        self.A = np.empty((level + 1, level + 1, level + 1), dtype=object)
        self.B = np.empty((level + 1, level + 1, level + 1), dtype=object)
        self.D = np.empty((level + 1, level + 1), dtype=object)
        self.Dxi = np.empty((level + 1, level + 1), dtype=object)
        self.Dxi2 = np.empty((level + 1, level + 1), dtype=object)

        self.DD = np.empty((level + 1, level + 1), dtype=object)
        self.D1 = np.empty((level + 1, level + 1), dtype=object)
        self.DT = np.empty((level + 1, level + 1, level + 1), dtype=object)

        for k in range(level + 1):
            self.phib[k] = self._phib(k)
            for i in range(level + 1):
                self.M[k, i] = self._M(k, i)
                self.D[k, i] = self._D(k, i)
                self.Dxi[k, i] = self._Dxi(k, i)
                self.Dxi2[k, i] = self._Dxi2(k, i)

                self.DD[k, i] = self._DD(k, i)
                self.D1[k, i] = self._D1(k, i)
                for j in range(level + 1):
                    self.A[k, i, j] = self._A(k, i, j)
                    self.B[k, i, j] = self._B(k, i, j)
                    self.DT[k, i, j] = self._DT(k, i, j)
            

    def compute_matrices(self, level):
        failed = True
        if self.use_cache:
            failed = self.load_cached_matrices()
        if failed or (not self.use_cache):
            self._compute_matrices(level)
            self.save_cached_matrices()

    def enforce_boundary_conditions_lsq(self, rhs=np.zeros(2), dim=1):
        level = len(self.basisfunctions.basis) - 1
        constraint_bottom = [self.basisfunctions.eval(i, 0.0) for i in range(level + 1)]
        constraint_top = [
            diff(self.basisfunctions.eval(i, z), z).subs(z, 1.0)
            for i in range(level + 1)
        ]
        A = Matrix([constraint_bottom, constraint_top])

        I = np.linspace(0, level, 1 + level, dtype=int)
        I_enforce = I[1:]
        rhs = np.zeros(2)
        # rhs = np.zeros(level)
        I_free = np.delete(I, I_enforce)
        A_enforce = A[:, list(I_enforce)]
        A_free = np.array(A[:, list(I_free)], dtype=float)
        AtA = A_enforce.T @ A_enforce
        reg = 10 ** (-6)
        A_enforce_inv = np.array((AtA + reg * np.eye(AtA.shape[0])).inv(), dtype=float)

        def f_1d(Q):
            for i, q in enumerate(Q.T):
                # alpha_enforce = q[I_enforce+1]
                alpha_free = q[I_free + 1]
                b = rhs - np.dot(A_free, alpha_free)
                # b = rhs
                result = np.dot(A_enforce_inv, A_enforce.T @ b)
                alpha = 1.0
                Q[I_enforce + 1, i] = (1 - alpha) * Q[I_enforce + 1, i] + (
                    alpha
                ) * result
            return Q

        def f_2d(Q):
            i1 = [[0] + [i + 1 for i in range(1 + level)]]
            i2 = [[0] + [i + 1 + 1 + level for i in range(1 + level)]]
            Q1 = Q[i1]
            Q2 = Q[i2]
            Q1 = f_1d(Q1)
            Q2 = f_1d(Q2)
            Q[i1] = Q1
            Q[i2] = Q2
            return Q

        if dim == 1:
            return f_1d
        elif dim == 2:
            return f_2d
        else:
            assert False

    def enforce_boundary_conditions_lsq2(self, rhs=np.zeros(2), dim=1):
        level = len(self.basisfunctions.basis) - 1
        constraint_bottom = [self.basisfunctions.eval(i, 0.0) for i in range(level + 1)]
        constraint_top = [
            diff(self.basisfunctions.eval(i, z), z).subs(z, 1.0)
            for i in range(level + 1)
        ]
        A = Matrix([constraint_bottom, constraint_top])

        I = np.linspace(0, level, 1 + level, dtype=int)
        I_enforce = I[1:]
        rhs = np.zeros(2)
        # rhs = np.zeros(level)
        I_free = np.delete(I, I_enforce)
        A_enforce = A[:, list(I_enforce)]
        A_free = np.array(A[:, list(I_free)], dtype=float)

        def obj(alpha0, lam):
            def f(alpha):
                return np.sum((alpha - alpha0) ** 2) + lam * np.sum(
                    np.array(np.dot(A, alpha) ** 2, dtype=float)
                )

            return f

        def f_1d(Q):
            for i, q in enumerate(Q.T):
                h = q[0]
                alpha = q[1:] / h
                f = obj(alpha, 0.1)
                result = lsq(f, alpha)
                Q[1:, i] = h * result.z
            return Q

        def f_2d(Q):
            i1 = [[0] + [i + 1 for i in range(1 + level)]]
            i2 = [[0] + [i + 1 + 1 + level for i in range(1 + level)]]
            Q1 = Q[i1]
            Q2 = Q[i2]
            Q1 = f_1d(Q1)
            Q2 = f_1d(Q2)
            Q[i1] = Q1
            Q[i2] = Q2
            return Q

        if dim == 1:
            return f_1d
        elif dim == 2:
            return f_2d
        else:
            assert False

    def enforce_boundary_conditions(
        self, enforced_basis=[-2, -1], rhs=np.zeros(2), dim=1
    ):
        level = len(self.basisfunctions.basis) - 1
        constraint_bottom = [self.basisfunctions.eval(i, 0.0) for i in range(level + 1)]
        constraint_top = [
            diff(self.basisfunctions.eval(i, z), z).subs(z, 1.0)
            for i in range(level + 1)
        ]
        A = Matrix([constraint_bottom, constraint_top][: len(enforced_basis)])

        # test to only constrain bottom
        # A = Matrix([constraint_bottom])
        # enforced_basis = [-1]
        # rhs=np.zeros(1)

        I = np.linspace(0, level, 1 + level, dtype=int)
        I_enforce = I[enforced_basis]
        I_free = np.delete(I, I_enforce)
        A_enforce = A[:, list(I_enforce)]
        A_free = np.array(A[:, list(I_free)], dtype=float)
        A_enforce_inv = np.array(A_enforce.inv(), dtype=float)

        def f_1d(Q):
            for i, q in enumerate(Q.T):
                alpha_enforce = q[I_enforce + 1]
                alpha_free = q[I_free + 1]
                b = rhs - np.dot(A_free, alpha_free)
                result = np.dot(A_enforce_inv, b)
                alpha = 1.0
                Q[I_enforce + 1, i] = (1 - alpha) * Q[I_enforce + 1, i] + (
                    alpha
                ) * result
            return Q

        def f_2d(Q):
            i1 = [[0] + [i + 1 for i in range(1 + level)]]
            i2 = [[0] + [i + 1 + 1 + level for i in range(1 + level)]]
            Q1 = Q[i1]
            Q2 = Q[i2]
            Q1 = f_1d(Q1)
            Q2 = f_1d(Q2)
            Q[i1] = Q1
            Q[i2] = Q2
            return Q

        if dim == 1:
            return f_1d
        elif dim == 2:
            return f_2d
        else:
            assert False
            
    """ 
    Compute phi_k(@xi=0)
    """

    def _phib(self, k):
        return self.basisfunctions.eval(k, self.basisfunctions.bounds()[0])

    """ 
    Compute <phi_k, phi_i>
    """

    def _M(self, k, i):
        return integrate(
            self.basisfunctions.weight(z) * self.basisfunctions.eval(k, z) * self.basisfunctions.eval(i, z), (z, self.basisfunctions.bounds()[0], self.basisfunctions.bounds()[1])
        )

    """ 
    Compute <phi_k, phi_i, phi_j>
    """

    def _A(self, k, i, j):
        return integrate(
            self.basisfunctions.weight(z) * self.basisfunctions.eval(k, z)
            * self.basisfunctions.eval(i, z)
            * self.basisfunctions.eval(j, z),
            (z, self.basisfunctions.bounds()[0], self.basisfunctions.bounds()[1]),
        )

    """ 
    Compute <(phi')_k, phi_j, int(phi)_j>
    """

    def _B(self, k, i, j):
        return integrate(
            self.basisfunctions.weight(z) * diff(self.basisfunctions.eval(k, z), z)
            * integrate(self.basisfunctions.eval(j, z), z)
            * self.basisfunctions.eval(i, z),
            (z, self.basisfunctions.bounds()[0], self.basisfunctions.bounds()[1]),
        )

    """ 
    Compute <(phi')_k, (phi')_j>
    """

    def _D(self, k, i):
        return integrate(
            self.basisfunctions.weight(z) * diff(self.basisfunctions.eval(k, z), z)
            * diff(self.basisfunctions.eval(i, z), z),
            (z, self.basisfunctions.bounds()[0], self.basisfunctions.bounds()[1]),
        )
        
    """ 
    Compute <(phi')_k, (phi')_j * xi>
    """
    def _Dxi(self, k, i):
        return integrate(
            self.basisfunctions.weight(z) * diff(self.basisfunctions.eval(k, z), z)
            * diff(self.basisfunctions.eval(i, z), z) * z,
            (z, self.basisfunctions.bounds()[0], self.basisfunctions.bounds()[1]),
        )
        """ 
    Compute <(phi')_k, (phi')_j * xi**2>
    """
    def _Dxi2(self, k, i):
        return integrate(
            self.basisfunctions.weight(z) * diff(self.basisfunctions.eval(k, z), z)
            * diff(self.basisfunctions.eval(i, z), z) * z * z,
            (z, self.basisfunctions.bounds()[0], self.basisfunctions.bounds()[1]),
        )

    """ 
    Compute <(phi)_k, (phi')_j>
    """

    def _D1(self, k, i):
        return integrate(
            self.basisfunctions.weight(z) * self.basisfunctions.eval(k, z) * diff(self.basisfunctions.eval(i, z), z),
            (z, self.basisfunctions.bounds()[0], self.basisfunctions.bounds()[1]),
        )

    """ 
    Compute <(phi)_k, (phi'')_j>
    """

    def _DD(self, k, i):
        return integrate(
            self.basisfunctions.weight(z) * self.basisfunctions.eval(k, z)
            * diff(diff(self.basisfunctions.eval(i, z), z), z),
           (z, self.basisfunctions.bounds()[0], self.basisfunctions.bounds()[1]),
        )

    """ 

    Compute <(phi')_k, (phi')_j>
    """

    def _DT(self, k, i, j):
        return integrate(
            self.basisfunctions.weight(z) * diff(self.basisfunctions.eval(k, z), z)
            * diff(self.basisfunctions.eval(i, z), z)
            * self.basisfunctions.eval(j, z),
            (z, self.basisfunctions.bounds()[0], self.basisfunctions.bounds()[1]),
        )


class BasisNoHOM(Basismatrices):
    def _A(self, k, i, j):
        count = 0
        # count += float(k > 0)
        count += float(i > 0)
        count += float(j > 0)
        # if count > 1:
        if (i == 0 and j == k) or (j == 0 and i == k) or (k == 0 and i == j):
            return super()._A(k, i, j)
        return 0

    def _B(self, k, i, j):
        count = 0
        # count += float(k > 0)
        count += float(i > 0)
        count += float(j > 0)
        # if count > 1:
        # if not (i==0 or j==0):
        if (i == 0 and j == k) or (j == 0 and i == k) or (k == 0 and i == j):
            return super()._B(k, i, j)
        return 0
        # return super()._B(k, i, j)



import numpy as np
import numpy.polynomial.legendre as L
import numpy.polynomial.chebyshev as C
from scipy.optimize import least_squares as lsq
import sympy
from sympy import Matrix, sqrt
from sympy.abc import x

from sympy import integrate, diff
from sympy import legendre
from sympy import lambdify, Rational


from library.zoomy_core.model.basemodel import (
    register_sympy_attribute,
    eigenvalue_dict_to_matrix,
)
from library.zoomy_core.model.basemodel import Model
import library.zoomy_core.model.initial_conditions as IC
import library.zoomy_core.model.boundary_conditions as BC

class GN(Model):
    def __init__(
        self,
        boundary_conditions= None,
        initial_conditions=None,
        dimension=1,
        fields=2,
        # D = h^3 / 3 * (dt * dx * u + u * dx^2 u - (dx u)^2)
        aux_variables=['dD_dx'],
        parameters={},
        _default_parameters={"g": 9.81},
        settings={},
        settings_default={},
    ):
        self.variables = register_sympy_attribute(fields, "q")
        self.n_variables = self.variables.length()
        super().__init__(
            dimension=dimension,
            fields=fields,
            aux_variables=aux_variables,
            parameters=parameters,
            _default_parameters=_default_parameters,
            boundary_conditions=boundary_conditions,
            initial_conditions=initial_conditions,
            settings={**settings_default, **settings},
        )
        
    def flux(self):
        fx = Matrix([0 for i in range(self.n_variables)])
        h = self.variables[0]
        hu = self.variables[1]

        param = self.parameters
        
        fx[0] = hu
        fx[1] = hu**2 / h + 1/2 * param.g * h**2 
        return [fx]

    

    def source_implicit(self):
        R = Matrix([0 for i in range(self.n_variables)])
        dD_dx = self.aux_variables.dD_dx


        R[0] = 0
        R[1] = dD_dx
        return R
        




def require(requirement):
    """
    Decorator to check if a requirement is met before executing the decorated function.

    Parameters:
    - requirement (str): The requirement string to evaluate. Should evaluate to True or False.

    Returns:
    - wrapper: The decorated function that will check the requirement before executing.
    """

    # decorator to check the assertion given in requirements given the settings
    def req_decorator(func):
        @wraps(func)
        def wrapper(settings, *args, **kwargs):
            requirement_evaluated = eval(requirement)
            if not requirement_evaluated:
                print("Requirement {}: {}".format(requirement, requirement_evaluated))
                assert requirement_evaluated
            return func(settings, *args, **kwargs)

        return wrapper

    return req_decorator



from library.zoomy_core.model.models.shallow_moments import (IC, Matrix, Model, eigenvalue_dict_to_matrix,
                                                  register_sympy_attribute, sympy)


class ShallowMomentsSediment(Model):
    """
    Shallow Moments Sediment 1d

    :gui:
    - tab: model
    - requires: [ 'mesh.dimension': 1 ]

    """

    def __init__(
        self,
        boundary_conditions,
        initial_conditions,
        aux_initial_conditions=IC.Constant(),
        dimension=1,
        fields=2,
        aux_variables=0,
        parameters={},
        _default_parameters={"g": 1.0, "ex": 0.0, "ez": 1.0},
        settings={},
        settings_default={"topography": False, "friction": []},
        basis=Basis(),
    ):
        self.basis = basis
        self.variables = register_sympy_attribute(fields, "q")
        self.n_variables = self.variables.length()
        self.level = self.n_variables - 2
        self.basis.compute_matrices(self.level)
        super().__init__(
            dimension=dimension,
            fields=fields,
            aux_variables=aux_variables,
            parameters=parameters,
            _default_parameters=_default_parameters,
            boundary_conditions=boundary_conditions,
            initial_conditions=initial_conditions,
            aux_initial_conditions=aux_initial_conditions,
            settings={**settings_default, **settings},
        )

    def flux(self):
        flux = Matrix([0 for i in range(self.n_variables)])
        h = self.variables[0]
        ha = self.variables[1 : 1 + self.level + 1]
        p = self.parameters
        flux[0] = ha[0]
        flux[1] = p.g * p.ez * h * h / 2
        for k in range(self.level + 1):
            for i in range(self.level + 1):
                for j in range(self.level + 1):
                    flux[k + 1] += (
                        ha[i] * ha[j] / h * self.basis.A[k, i, j] / self.basis.M[k, k]
                    )
        return [flux]

    def nonconservative_matrix(self):
        nc = Matrix([[0 for i in range(self.n_variables)] for j in range(self.n_variables)])
        h = self.variables[0]
        ha = self.variables[1 : 1 + self.level + 1]
        p = self.parameters
        um = ha[0] / h
        # nc[1, 0] = - p.g * p.ez * h
        for k in range(1, self.level + 1):
            nc[k + 1, k + 1] += um
        for k in range(self.level + 1):
            for i in range(1, self.level + 1):
                for j in range(1, self.level + 1):
                    nc[k + 1, i + 1] -= (
                        ha[j] / h * self.basis.B[k, i, j] / self.basis.M[k, k]
                    )
        return [nc]

    def eigenvalues(self):
        A = self.normal[0] * self.sympy_quasilinear_matrix[0]
        for d in range(1, self.dimension):
            A += self.normal[d] * self.sympy_quasilinear_matrix[d]
        alpha_erase = self.variables[2:]
        for alpha_i in alpha_erase:
            A = A.subs(alpha_i, 0)
        return eigenvalue_dict_to_matrix(A.eigenvals())

    def source(self):
        out = Matrix([0 for i in range(self.n_variables)])
        return out

    def topography(self):
        assert "dhdx" in vars(self.aux_variables)
        out = Matrix([0 for i in range(self.n_variables)])
        h = self.variables[0]
        p = self.parameters
        dhdx = self.aux_variables.dhdx
        out[1] = h * p.g * (p.ex - p.ez * dhdx)
        return out

    def inclined_plane(self):
        out = Matrix([0 for i in range(self.n_variables)])
        h = self.variables[0]
        p = self.parameters
        out[1] = h * p.g * (p.ex)
        return out

    def newtonian(self):
        """
        :gui:
            - requires_parameter: ('nu', 0.0)
        """
        assert "nu" in vars(self.parameters)
        out = Matrix([0 for i in range(self.n_variables)])
        h = self.variables[0]
        ha = self.variables[1 : 1 + self.level + 1]
        p = self.parameters
        for k in range(1 + self.level):
            for i in range(1 + self.level):
                out[1 + k] += (
                    -p.nu / h * ha[i] / h * self.basis.D[i, k] / self.basis.M[k, k]
                )
        return out

    def slip(self):
        """
        :gui:
            - requires_parameter: ('lamda', 0.0)
            - requires_parameter: ('rho', 1.0)
        """
        assert "lamda" in vars(self.parameters)
        assert "rho" in vars(self.parameters)
        out = Matrix([0 for i in range(self.n_variables)])
        h = self.variables[0]
        ha = self.variables[1 : 1 + self.level + 1]
        p = self.parameters
        for k in range(1 + self.level):
            for i in range(1 + self.level):
                out[1 + k] += -1.0 / p.lamda / p.rho * ha[i] / h / self.basis.M[k, k]
        return out

    def chezy(self):
        """
        :gui:
            - requires_parameter: ('C', 1000.0)
        """
        assert "C" in vars(self.parameters)
        out = Matrix([0 for i in range(self.n_variables)])
        h = self.variables[0]
        ha = self.variables[1 : 1 + self.level + 1]
        p = self.parameters
        tmp = 0
        for i in range(1 + self.level):
            for j in range(1 + self.level):
                tmp += ha[i] * ha[j] / h / h
        sqrt = sympy.sqrt(tmp)
        for k in range(1 + self.level):
            for l in range(1 + self.level):
                out[1 + k] += -1.0 / (p.C**2 * self.basis.M[k, k]) * ha[l] * sqrt / h
        return out

    def chezy_ssf(self):
        """
        :gui:
            - requires_parameter: ('C', 1000.0)
        """
        assert "Cf" in vars(self.parameters)
        out = Matrix([0 for i in range(self.n_variables)])
        h = self.variables[0]
        ha = self.variables[1 : 1 + self.level + 1]
        p = self.parameters
        tmp = 0
        for i in range(1 + self.level):
            for j in range(1 + self.level):
                tmp += ha[i] * ha[j] / h / h
        sqrt = sympy.sqrt(tmp)
        for k in range(1 + self.level):
            for l in range(1 + self.level):
                out[1 + k] += -(p.Cf * self.basis.M[k, k]) * ha[l] * sqrt / h
        return out


class ShallowMomentsSediment2d(Model):
    def __init__(
        self,
        boundary_conditions,
        initial_conditions,
        dimension=2,
        fields=3,
        aux_variables=0,
        parameters={},
        _default_parameters={"g": 1.0, "ex": 0.0, "ey": 0.0, "ez": 1.0},
        settings={},
        settings_default={"topography": False, "friction": []},
        basis=Basis(),
    ):
        self.basis = basis
        self.variables = register_sympy_attribute(fields, "q")
        self.n_variables = self.variables.length()
        self.level = int((self.n_variables - 1) / 2) - 1
        self.basis.compute_matrices(self.level)
        super().__init__(
            dimension=dimension,
            fields=fields,
            aux_variables=aux_variables,
            parameters=parameters,
            _default_parameters=_default_parameters,
            boundary_conditions=boundary_conditions,
            initial_conditions=initial_conditions,
            settings={**settings_default, **settings},
        )

    def flux(self):
        offset = self.level + 1
        flux_x = Matrix([0 for i in range(self.n_variables)])
        flux_y = Matrix([0 for i in range(self.n_variables)])
        h = self.variables[0]
        ha = self.variables[1 : 1 + self.level + 1]
        hb = self.variables[1 + self.level + 1 : 1 + 2 * (self.level + 1)]
        p = self.parameters
        flux_x[0] = ha[0]
        flux_x[1] = p.g * p.ez * h * h / 2
        for k in range(self.level + 1):
            for i in range(self.level + 1):
                for j in range(self.level + 1):
                    # TODO avoid devision by zero
                    flux_x[k + 1] += (
                        ha[i] * ha[j] / h * self.basis.A[k, i, j] / self.basis.M[k, k]
                    )
        for k in range(self.level + 1):
            for i in range(self.level + 1):
                for j in range(self.level + 1):
                    # TODO avoid devision by zero
                    flux_x[k + 1 + offset] += (
                        hb[i] * ha[j] / h * self.basis.A[k, i, j] / self.basis.M[k, k]
                    )

        flux_y[0] = hb[0]
        flux_y[1 + offset] = p.g * p.ez * h * h / 2
        for k in range(self.level + 1):
            for i in range(self.level + 1):
                for j in range(self.level + 1):
                    # TODO avoid devision by zero
                    flux_y[k + 1] += (
                        hb[i] * ha[j] / h * self.basis.A[k, i, j] / self.basis.M[k, k]
                    )
        for k in range(self.level + 1):
            for i in range(self.level + 1):
                for j in range(self.level + 1):
                    # TODO avoid devision by zero
                    flux_y[k + 1 + offset] += (
                        hb[i] * hb[j] / h * self.basis.A[k, i, j] / self.basis.M[k, k]
                    )
        return [flux_x, flux_y]

    def nonconservative_matrix(self):
        offset = self.level + 1
        nc_x = Matrix([[0 for i in range(self.n_variables)] for j in range(self.n_variables)])
        nc_y = Matrix([[0 for i in range(self.n_variables)] for j in range(self.n_variables)])
        h = self.variables[0]
        ha = self.variables[1 : 1 + self.level + 1]
        hb = self.variables[1 + offset : 1 + offset + self.level + 1]
        p = self.parameters
        um = ha[0] / h
        vm = hb[0] / h
        for k in range(1, self.level + 1):
            nc_x[k + 1, k + 1] += um
            nc_y[k + 1, k + 1 + offset] += um
        for k in range(self.level + 1):
            for i in range(1, self.level + 1):
                for j in range(1, self.level + 1):
                    nc_x[k + 1, i + 1] -= (
                        ha[j] / h * self.basis.B[k, i, j] / self.basis.M[k, k]
                    )
                    nc_y[k + 1, i + 1 + offset] -= (
                        ha[j] / h * self.basis.B[k, i, j] / self.basis.M[k, k]
                    )

        for k in range(1, self.level + 1):
            nc_x[k + 1 + offset, k + 1] += vm
            nc_y[k + 1 + offset, k + 1 + offset] += vm
        for k in range(self.level + 1):
            for i in range(1, self.level + 1):
                for j in range(1, self.level + 1):
                    nc_x[k + 1 + offset, i + 1] -= (
                        hb[j] / h * self.basis.B[k, i, j] / self.basis.M[k, k]
                    )
                    nc_y[k + 1 + offset, i + 1 + offset] -= (
                        hb[j] / h * self.basis.B[k, i, j] / self.basis.M[k, k]
                    )
        return [nc_x, nc_y]

    def eigenvalues(self):
        # we delete heigher order moments (level >= 2) for analytical eigenvalues
        offset = self.level + 1
        A = self.normal[0] * self.sympy_quasilinear_matrix[0]
        for d in range(1, self.dimension):
            A += self.normal[d] * self.sympy_quasilinear_matrix[d]
        alpha_erase = self.variables[2 : 2 + self.level]
        beta_erase = self.variables[2 + offset : 2 + offset + self.level]
        for alpha_i in alpha_erase:
            A = A.subs(alpha_i, 0)
        for beta_i in beta_erase:
            A = A.subs(beta_i, 0)
        return eigenvalue_dict_to_matrix(A.eigenvals())

    def source(self):
        out = Matrix([0 for i in range(self.n_variables)])
        return out

    def topography(self):
        assert "dhdx" in vars(self.aux_variables)
        assert "dhdy" in vars(self.aux_variables)
        offset = self.level + 1
        out = Matrix([0 for i in range(self.n_variables)])
        h = self.variables[0]
        p = self.parameters
        dhdx = self.aux_variables.dhdx
        dhdy = self.aux_variables.dhdy
        out[1] = h * p.g * (p.ex - p.ez * dhdx)
        out[1 + offset] = h * p.g * (p.ey - p.ez * dhdy)
        return out

    def newtonian(self):
        assert "nu" in vars(self.parameters)
        out = Matrix([0 for i in range(self.n_variables)])
        offset = self.level + 1
        h = self.variables[0]
        ha = self.variables[1 : 1 + self.level + 1]
        hb = self.variables[1 + offset : 1 + self.level + 1 + offset]
        p = self.parameters
        for k in range(1 + self.level):
            for i in range(1 + self.level):
                out[1 + k] += (
                    -p.nu / h * ha[i] / h * self.basis.D[i, k] / self.basis.M[k, k]
                )
                out[1 + k + offset] += (
                    -p.nu / h * hb[i] / h * self.basis.D[i, k] / self.basis.M[k, k]
                )
        return out

    def slip(self):
        assert "lamda" in vars(self.parameters)
        assert "rho" in vars(self.parameters)
        out = Matrix([0 for i in range(self.n_variables)])
        offset = self.level + 1
        h = self.variables[0]
        h = self.variables[0]
        ha = self.variables[1 : 1 + self.level + 1]
        hb = self.variables[1 + offset : 1 + self.level + 1 + offset]
        p = self.parameters
        for k in range(1 + self.level):
            for i in range(1 + self.level):
                out[1 + k] += -1.0 / p.lamda / p.rho * ha[i] / h / self.basis.M[k, k]
                out[1 + k + offset] += (
                    -1.0 / p.lamda / p.rho * hb[i] / h / self.basis.M[k, k]
                )
        return out

    def chezy(self):
        assert "C" in vars(self.parameters)
        out = Matrix([0 for i in range(self.n_variables)])
        offset = self.level + 1
        h = self.variables[0]
        ha = self.variables[1 : 1 + self.level + 1]
        hb = self.variables[1 + offset : 1 + self.level + 1 + offset]
        p = self.parameters
        tmp = 0
        for i in range(1 + self.level):
            for j in range(1 + self.level):
                tmp += ha[i] * ha[j] / h / h + hb[i] * hb[j] / h / h
        sqrt = sympy.sqrt(tmp)
        for k in range(1 + self.level):
            for l in range(1 + self.level):
                out[1 + k] += -1.0 / (p.C**2 * self.basis.M[k, k]) * ha[l] * sqrt / h
                out[1 + k + offset] += (
                    -1.0 / (p.C**2 * self.basis.M[k, k]) * hb[l] * sqrt / h
                )
        return out



import numpy as np
from attr import define
from typing import Callable, Optional

from library.zoomy_core.misc.custom_types import FArray
from library.zoomy_core.mesh.mesh import Mesh
import library.zoomy_core.misc.io as io
import library.zoomy_core.misc.interpolation as interpolate_mesh

# from library.zoomy_core.mesh import Mesh2D


@define(slots=True, frozen=True)
class InitialConditions:
    def apply(self, X, Q):
        assert False
        return Q


@define(slots=True, frozen=True)
class Constant(InitialConditions):
    constants: Callable[[int], FArray] = lambda n_variables: np.array(
        [1.0] + [0.0 for i in range(n_variables - 1)]
    )

    def apply(self, X, Q):
        n_variables = Q.shape[0]
        for i in range(Q.shape[1]):
            Q[:, i] = self.constants(n_variables)
        return Q


@define(slots=True, frozen=False)
class RP(InitialConditions):
    low: Callable[[int], FArray] = lambda n_variables: np.array(
        [1.0 * (i == 0) for i in range(n_variables)]
    )
    high: Callable[[int], FArray] = lambda n_variables: np.array(
        [2.0 * (i == 0) for i in range(n_variables)]
    )
    jump_position_x: float = 0.0

    def apply(self, X, Q):
        assert X.shape[1] == Q.shape[1]
        n_variables = Q.shape[0]
        for i in range(Q.shape[1]):
            if X[0, i] < self.jump_position_x:
                Q[:, i] = self.high(n_variables)
            else:
                Q[:, i] = self.low(n_variables)
        return Q


@define(slots=True, frozen=False)
class RP2d(InitialConditions):
    low: Callable[[int], FArray] = lambda n_variables: np.array(
        [1.0 * (i == 0) for i in range(n_variables)]
    )
    high: Callable[[int], FArray] = lambda n_variables: np.array(
        [2.0 * (i == 0) for i in range(n_variables)]
    )
    jump_position_x: float = 0.0
    jump_position_y: float = 0.0

    def apply(self, X, Q):
        assert X.shape[1] == Q.shape[1]
        n_variables = Q.shape[0]
        for i in range(Q.shape[1]):
            if X[0, i] < self.jump_position_x and X[1, i] < self.jump_position_y:
                Q[:, i] = self.high(n_variables)
            else:
                Q[:, i] = self.low(n_variables)
        return Q


@define(slots=True, frozen=False)
class RP3d(InitialConditions):
    low: Callable[[int], FArray] = lambda n_variables: np.array(
        [1.0 * (i == 0) for i in range(n_variables)]
    )
    high: Callable[[int], FArray] = lambda n_variables: np.array(
        [2.0 * (i == 0) for i in range(n_variables)]
    )
    jump_position_x: float = 0.0
    jump_position_y: float = 0.0
    jump_position_z: float = 0.0

    def apply(self, X, Q):
        assert X.shape[1] == Q.shape[1]
        n_variables = Q.shape[0]
        for i in range(Q.shape[1]):
            if (
                X[0, i] < self.jump_position_x
                and X[1, i] < self.jump_position_y
                and X[2, i] < self.jump_position_z
            ):
                Q[:, i] = self.high(n_variables)
            else:
                Q[:, i] = self.low(n_variables)
        return Q


@define(slots=True, frozen=False)
class RadialDambreak(InitialConditions):
    low: Callable[[int], FArray] = lambda n_variables: np.array(
        [1.0 * (i == 0) for i in range(n_variables)]
    )
    high: Callable[[int], FArray] = lambda n_variables: np.array(
        [2.0 * (i == 0) for i in range(n_variables)]
    )
    radius: float = 0.1

    def apply(self, X, Q):
        dim = X.shape[0]
        center = np.zeros(dim)
        for d in range(dim):
            center[d] = X[d, :].mean()
        assert X.shape[1] == Q.shape[1]
        n_variables = Q.shape[0]
        for i in range(Q.shape[1]):
            if np.linalg.norm(X[:, i] - center) <= self.radius:
                Q[:, i] = self.high(n_variables)
            else:
                Q[:, i] = self.low(n_variables)
        return Q


@define(slots=True, frozen=True)
class UserFunction(InitialConditions):
    function: Optional[Callable[[FArray], FArray]] = None

    def apply(self, X, Q):
        assert X.shape[1] == Q.shape[1]
        if self.function is None:
            self.function = lambda x: np.zeros(Q.shape[1])
        for i, x in enumerate(X.T):
            Q[:, i] = self.function(x)
        return Q


# TODO do time interpolation
@define(slots=True, frozen=True)
class RestartFromHdf5(InitialConditions):
    path_to_fields: Optional[str] = None
    mesh_new: Optional[Mesh] = None
    mesh_identical: bool = False
    path_to_old_mesh: Optional[str] = None
    snapshot: Optional[int] = -1
    map_fields: Optional[dict] = None

    def apply(self, X, Q):
        assert self.mesh_new is not None
        assert self.path_to_fields is not None
        assert X.shape[0] == Q.shape[0]
        if self.map_fields is None:
            map_fields = {i: i for i in range(Q.shape[1])}
        else:
            map_fields = self.map_fields
        mesh = Mesh.from_hdf5(self.path_to_old_mesh)
        _Q, _Qaux, time = io.load_fields_from_hdf5(
            self.path_to_fields, i_snapshot=self.snapshot
        )
        Q = np.zeros_like(Q)
        if self.mesh_identical:
            Q[:, list(map_fields.values())] = _Q[:, list(map_fields.keys())]
        else:
            assert self.path_to_old_mesh is not None
            Q[:, list(map_fields.values())] = interpolate_mesh.to_new_mesh(
                _Q, mesh, self.mesh_new
            )[:, list(map_fields.keys())]
        return Q



import os
import numpy as np

# import scipy.interpolate as interp
# from functools import wraps

from attr import define
from typing import Callable, Optional, Any
from types import SimpleNamespace

from sympy import MatrixSymbol
from sympy import MutableDenseNDimArray as ZArray

from library.zoomy_core.misc.custom_types import FArray
from library.zoomy_core.misc.static_class import register_static_pytree
from library.zoomy_core.misc.logger_config import logger






@register_static_pytree
@define(slots=True, frozen=False, kw_only=True)
class Zstruct(SimpleNamespace):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def __getitem__(self, key):
        return self.values()[key]

    def length(self):
        return len(self.values())

    def get_list(self, recursive: bool = True):
        if recursive:
            output = []
            for item in self.values():
                if hasattr(item, 'get_list'):
                    # If item is a Zstruct or similar, call get_list recursively
                    output.append(item.get_list(recursive=True))
                else:
                    output.append(item)
            return output
        else:
            return self.values()
        
    def as_dict(self, recursive: bool = True):
        if recursive:
            output = {}
            for key, value in self.items():
                if hasattr(value, 'as_dict'):
                    # If value is a Zstruct or similar, call as_dict recursively
                    output[key] = value.as_dict(recursive=True)
                else:
                    output[key] = value
            return output
        else:
            return self.__dict__

    
    def items(self, resursive: bool = False):
        return self.as_dict(recursive=resursive).items()
    
    def keys(self):
        return list(self.as_dict(recursive=False).keys())
    
    def values(self):
        return list(self.as_dict(recursive=False).values())
    
    def contains(self, key):
        if self.as_dict(recursive=False).get(key) is not None:
            return True
        return False
    
    def update(self, zstruct, recursive: bool = True):
        """
        Update the current Zstruct with another Zstruct or dictionary.
        """
        if not isinstance(zstruct, Zstruct):
            raise TypeError("zstruct must be a Zstruct or a dictionary.")
        
        if recursive:
            # Update each attribute recursively
            for key, value in zstruct.as_dict(recursive=False).items():
                if hasattr(self, key):
                    current_value = getattr(self, key)
                    if isinstance(current_value, Zstruct) and isinstance(value, Zstruct):
                        # If both are Zstructs, update recursively
                        current_value.update(value, recursive=True)
                    else:
                        setattr(self, key, value)
                else:
                    setattr(self, key, value)
        else:
            # Update only the top-level attributes
            for key, value in zstruct.as_dict(recursive=False).items():
                setattr(self, key, value)
                
    @classmethod
    def from_dict(cls, d):
        """
        Create a Zstruct recursively from a dictionary.
        
        Args:
            d (dict): Dictionary.
        
        Returns:
            Zstruct: An instance of Zstruct.
        """
        if not isinstance(d, dict):
            raise TypeError("Input must be a dictionary.")
        
        # Convert the dictionary to a Zstruct
        for k, v in d.items():
            if isinstance(v, dict):
                d[k] = Zstruct.from_dict(v)     
        return cls(**d)

    
@register_static_pytree
@define(slots=True, frozen=False, kw_only=True)
class Settings(Zstruct):
    """
    Settings class for the application.
    
    Args: 
        **kwargs: Arbitrary keyword arguments to set as attributes.
        
    Returns:
        An `IterableNamespace` instance.
    """
    
    def __init__(self, **kwargs):
        # assert that kwargs constains name
        if 'output' not in kwargs or not isinstance(kwargs['output'], Zstruct):
            logger.warning("No 'output' Zstruct found in Settings. Default: Zstruct(directory='output', filename='simulation', clean_directory=False)")
            kwargs['output'] = Zstruct(directory='output', filename='simulation', clean_directory=True)
        output = kwargs['output']
        if not output.contains('directory'):
            logger.warning("No 'directory' attribute found in output Zstruct. Default: 'output'")
            kwargs['output'] = Zstruct(directory='output', **output.as_dict())
        if not output.contains('filename'):
            logger.warning("No 'filename' attribute found in output Zstruct. Default: 'simulation'")
            kwargs['output'] = Zstruct(filename='simulation', **output.as_dict())
        if not output.contains('clean_directory'):
            logger.warning("No 'clean_directory' attribute found in output Zstruct. Default: False")
            kwargs['output'] = Zstruct(clean_directory=False, **output.as_dict())
        super().__init__(**kwargs)
        
    @classmethod
    def default(cls):
        """
        Returns a default Settings instance.
        """
        return cls(
            output=Zstruct(directory='output', filename='simulation', clean_directory=False)
        )
    

def compute_transverse_direction(normal):
    dim = normal.shape[0]
    if dim == 1:
        return np.zeros_like(normal)
    elif dim == 2:
        transverse = np.zeros((2), dtype=float)
        transverse[0] = -normal[1]
        transverse[1] = normal[0]
        return transverse
    elif dim == 3:
        cartesian_x = np.array([1, 0, 0], dtype=float)
        transverse = np.cross(normal, cartesian_x)
        return transverse
    else:
        assert False


def extract_momentum_fields_as_vectors(Q, momentum_fields, dim):
    num_fields = len(momentum_fields)
    num_momentum_eqns = int(num_fields / dim)
    Qnew = np.empty((num_momentum_eqns, dim))
    for i_eq in range(num_momentum_eqns):
        for i_dim in range(dim):
            Qnew[i_eq, i_dim] = Q[momentum_fields[i_dim * num_momentum_eqns + i_eq]]
    return Qnew


def projection_in_normal_and_transverse_direction(Q, momentum_fields, normal):
    dim = normal.shape[0]
    transverse_directions = compute_transverse_direction(normal)
    Q_momentum_eqns = extract_momentum_fields_as_vectors(Q, momentum_fields, dim)
    Q_normal = np.zeros((Q_momentum_eqns.shape[0]), dtype=float)
    Q_transverse = np.zeros((Q_momentum_eqns.shape[0]), dtype=float)
    for d in range(dim):
        Q_normal += Q_momentum_eqns[:, d] * normal[d]
        Q_transverse += Q_momentum_eqns[:, d] * transverse_directions[d]
    return Q_normal, Q_transverse


def projection_in_x_y_direction(Qn, Qt, normal):
    dim = normal.shape[0]
    num_momentum_fields = Qn.shape[0]
    transverse_directions = compute_transverse_direction(normal)
    Q = np.empty((num_momentum_fields * dim), dtype=float)
    for i in range(num_momentum_fields):
        for d in range(dim):
            Q[i + d * num_momentum_fields] = (
                Qn[i] * normal[d] + Qt[i] * transverse_directions[d]
            )
    return Q


def project_in_x_y_and_recreate_Q(Qn, Qt, Qorig, momentum_eqns, normal):
    Qnew = np.array(Qorig)
    Qnew[momentum_eqns] = projection_in_x_y_direction(Qn, Qt, normal)
    return Qnew



import numpy as np
import numpy.polynomial.legendre as L
import numpy.polynomial.chebyshev as C
from scipy.optimize import least_squares as lsq
import sympy
from sympy import Matrix, Piecewise
from sympy.abc import x

from sympy import integrate, diff
from sympy import legendre
from sympy import lambdify

from attrs import define, field
import attr
from typing import Union, Dict, List


from library.zoomy_core.model.basemodel import (
    register_sympy_attribute,
    eigenvalue_dict_to_matrix,
)
from library.zoomy_core.model.basemodel import Model
from library.zoomy_core.model.models.basismatrices import Basismatrices
from library.zoomy_core.model.models.basisfunctions import Legendre_shifted, Basisfunction


    

@define(frozen=True, slots=True, kw_only=True)
class ShallowMomentsTopo(Model):
    dimension: int = 2
    level: int
    variables: Union[list, int] = field(init=False)
    positive_variables: Union[List[int], Dict[str, int], None] = attr.ib(default=attr.Factory(lambda: [1]))    
    aux_variables: Union[list, int] = field(default=0)
    basisfunctions: Union[Basisfunction, type[Basisfunction]] = field(default=Legendre_shifted)
    basismatrices: Basismatrices = field(init=False)

    _default_parameters: dict = field(
        init=False,
        factory=lambda: {"g": 9.81, "ex": 0.0, "ey": 0.0, "ez": 1.0, "eps_low_water": 1e-6, "rho": 1000},
    )

    def __attrs_post_init__(self):
        object.__setattr__(self, "variables", ((self.level+1)*self.dimension)+2)
        object.__setattr__(self, "aux_variables", 2*((self.level+1)*self.dimension+2))
        super().__attrs_post_init__()
        aux_variables = self.aux_variables
        aux_var_list = aux_variables.keys()
        object.__setattr__(self, "aux_variables", register_sympy_attribute(aux_var_list, "qaux_"))

        # Recompute basis matrices
        object.__setattr__(self, "basisfunctions", self.basisfunctions(level=self.level))
        basismatrices = Basismatrices(self.basisfunctions)
        basismatrices.compute_matrices(self.level)
        object.__setattr__(self, "basismatrices", basismatrices)

    def get_primitives(self):
        offset = self.level + 1
        b = self.variables[0]
        h = self.variables[1]
        hinv = 1/h
        ha = self.variables[2 : 2 + self.level + 1]
        alpha = [ha[i] * hinv for i in range(offset)]
        if self.dimension == 1:
            hb = [0 for i in range(self.level+1)]
        else:
            hb = self.variables[2 + offset : 2 + offset + self.level + 1]
        beta = [hb[i] * hinv for i in range(offset)]
        return [b, h, alpha, beta, hinv]


    def project_2d_to_3d(self):
        out = Matrix([0 for i in range(6)])
        level = self.level
        offset = level+1
        offset_aux = self.n_variables
        x = self.position[0]
        y = self.position[1]
        z = self.position[2]
        
        b, h, alpha, beta, hinv = self.get_primitives()
        dbdx = self.aux_variables[0]
        dhdx = self.aux_variables[1]
        dbdy = self.aux_variables[offset_aux]
        dhdy = self.aux_variables[1+offset_aux]
        dalphadx = [self.aux_variables[2+i] for i in range(offset)]
        if self.dimension == 2:
            dbetady = [self.aux_variables[2+i+offset_aux] for i in range(offset)]
        
        psi = [self.basisfunctions.eval_psi(k, z) for k in range(level+1)]
        phi = [self.basisfunctions.eval(k, z) for k in range(level+1)]

        rho_w = 1000.
        g = 9.81
        u_3d = self.basismatrices.basisfunctions.reconstruct_velocity_profile_at(alpha, z)
        v_3d = 0
        def dot(a, b):
            s = 0
            for i in range(len(a)):
                s += a[i] * b[i]
            return s
        w_3d = - dhdx * dot(alpha,psi) - h * dot(dalphadx,psi) + dot(alpha, phi) * (z * dhdx + dbdx)
        if self.dimension == 2:
            v_3d = self.basismatrices.basisfunctions.reconstruct_velocity_profile_at(beta, z)
            w_3d += - dhdy * dot(beta,psi) - h * dot(dbetady,psi) + dot(beta, phi) * (z * dhdy + dbdy)

        out[0] = b
        out[1] = h
        out[2] = u_3d
        out[3] = v_3d
        out[4] = w_3d
        out[5] = rho_w * g * h * (1-z)
        return out

    def flux(self):
        flux_x = Matrix([0 for i in range(self.n_variables)])
        flux_y = Matrix([0 for i in range(self.n_variables)])
        b, h, alpha, beta, hinv = self.get_primitives()
        p = self.parameters
        flux_x[1] = h * alpha[0]
        flux_x[2] = p.g * p.ez * h * h / 2
        for k in range(self.level + 1):
            for i in range(self.level + 1):
                for j in range(self.level + 1):
                    flux_x[k + 2] += (
                        h * alpha[i] * alpha[j]
                        * self.basismatrices.A[k, i, j]
                        / self.basismatrices.M[k, k]
                    )
        if self.dimension == 2:
            offset = self.level + 1
            p = self.parameters
            for k in range(self.level + 1):
                for i in range(self.level + 1):
                    for j in range(self.level + 1):
                        flux_x[1 + k + 1 + offset] += (
                            h * beta[i] * alpha[j]
                            * self.basismatrices.A[k, i, j]
                            / self.basismatrices.M[k, k]
                        )

            flux_y[1] = h * beta[0]
            flux_y[2 + offset] = p.g * p.ez * h * h / 2
            for k in range(self.level + 1):
                for i in range(self.level + 1):
                    for j in range(self.level + 1):
                        flux_y[1 + k + 1] += (
                            h * beta[i] * alpha[j]
                            * self.basismatrices.A[k, i, j]
                            / self.basismatrices.M[k, k]
                        )
            for k in range(self.level + 1):
                for i in range(self.level + 1):
                    for j in range(self.level + 1):
                        flux_y[1 + k + 1 + offset] += (
                            h * beta[i] * beta[j]
                            * self.basismatrices.A[k, i, j]
                            / self.basismatrices.M[k, k]
                        )
        return [flux_x, flux_y][:self.dimension]

    def nonconservative_matrix(self):
        nc_x = Matrix([[0 for i in range(self.n_variables)] for j in range(self.n_variables)])
        nc_y = Matrix([[0 for i in range(self.n_variables)] for j in range(self.n_variables)])
        b, h, alpha, beta, hinv = self.get_primitives()
        p = self.parameters
        um = alpha[0]
        for k in range(1, self.level + 1):
            nc_x[1+k + 1, 1+k + 1] += um
        for k in range(self.level + 1):
            for i in range(1, self.level + 1):
                for j in range(1, self.level + 1):
                    nc_x[1+k + 1, 1+i + 1] -= (
                        alpha[j]
                        * self.basismatrices.B[k, i, j]
                        / self.basismatrices.M[k, k]
                    )
        if self.dimension == 2:
            offset = self.level + 1
            b, h, alpha, beta, hinv = self.get_primitives()
            p = self.parameters
            um = alpha[0]
            vm = beta[0]
            for k in range(1, self.level + 1):
                nc_y[1+k + 1, 1+k + 1 + offset] += um
            for k in range(self.level + 1):
                for i in range(1, self.level + 1):
                    for j in range(1, self.level + 1):
                        nc_x[1+k + 1, 1+i + 1] -= (
                            alpha[j]
                            * self.basismatrices.B[k, i, j]
                            / self.basismatrices.M[k, k]
                        )
                        nc_y[1+k + 1, 1+i + 1 + offset] -= (
                            alpha[j]
                            * self.basismatrices.B[k, i, j]
                            / self.basismatrices.M[k, k]
                        )

            for k in range(1, self.level + 1):
                nc_x[1+k + 1 + offset, 1+k + 1] += vm
                nc_y[1+k + 1 + offset, 1+k + 1 + offset] += vm
            for k in range(self.level + 1):
                for i in range(1, self.level + 1):
                    for j in range(1, self.level + 1):
                        nc_x[1+k + 1 + offset, 1+i + 1] -= (
                            beta[j]
                            * self.basismatrices.B[k, i, j]
                            / self.basismatrices.M[k, k]
                        )
                        nc_y[1+k + 1 + offset, 1+i + 1 + offset] -= (
                            beta[j]
                            * self.basismatrices.B[k, i, j]
                            / self.basismatrices.M[k, k]
                        )
        return [-nc_x, -nc_y][:self.dimension]

    def eigenvalues(self):
        # we delete heigher order moments (level >= 2) for analytical eigenvalues
        offset = self.level + 1
        A = self.normal[0] * self.quasilinear_matrix()[0]
        for d in range(1, self.dimension):
            A += self.normal[d] * self.quasilinear_matrix()[d]
        b, h, alpha, beta, hinv = self.get_primitives()
        alpha_erase = alpha[1:] if self.level >= 2 else []
        beta_erase = beta[1:] if self.level >= 2 else []
        for alpha_i in alpha_erase:
            A = A.subs(alpha_i, 0)
        for beta_i in beta_erase:
            A = A.subs(beta_i, 0)
        return eigenvalue_dict_to_matrix(A.eigenvals())

    def source(self):
        out = Matrix([0 for i in range(self.n_variables)])
        return out

    def newtonian(self):
        assert "nu" in vars(self.parameters)
        out = Matrix([0 for i in range(self.n_variables)])
        offset = self.level + 1
        b, h, alpha, beta, hinv = self.get_primitives()
        p = self.parameters
        for k in range(1 + self.level):
            for i in range(1 + self.level):
                out[1+1 + k] += (
                    -p.nu
                    * alpha[i]
                    * hinv
                    * self.basismatrices.D[i, k]
                    / self.basismatrices.M[k, k]
                )
                if self.dimension == 2:
                    out[1+1 + k + offset] += (
                        -p.nu
                        * beta[i]
                        * hinv
                        * self.basismatrices.D[i, k]
                        / self.basismatrices.M[k, k]
                    )
        return out

    def slip_mod(self):
        assert "lamda" in vars(self.parameters)
        assert "rho" in vars(self.parameters)
        assert "c_slipmod" in vars(self.parameters)

        out = Matrix([0 for i in range(self.n_variables)])
        offset = self.level+1
        b, h, alpha, beta, hinv = self.get_primitives()
        p = self.parameters
        ub = 0
        vb = 0
        for i in range(1 + self.level):
            ub += alpha[i]
            vb += beta[i]
        for k in range(1, 1 + self.level):
            out[2 + k] += (
                -1.0 * p.c_slipmod / p.lamda / p.rho * ub / self.basismatrices.M[k, k]
            )
            if self.dimension == 2:
                out[2+offset+k] += (
                    -1.0 * p.c_slipmod / p.lamda / p.rho * vb / self.basismatrices.M[k, k]
                )
        return out


    def slip(self):
        assert "lamda" in vars(self.parameters)
        assert "rho" in vars(self.parameters)
        out = Matrix([0 for i in range(self.n_variables)])
        offset = self.level + 1
        b, h, alpha, beta, hinv = self.get_primitives()
        p = self.parameters
        for k in range(1 + self.level):
            for i in range(1 + self.level):
                out[1+1 + k] += (
                    -1.0 / p.lamda / p.rho * alpha[i] / self.basismatrices.M[k, k]
                )
                if self.dimension == 2:
                    out[1+1 + k + offset] += (
                        -1.0 / p.lamda / p.rho * beta[i] / self.basismatrices.M[k, k]
                    )
        return out

    def chezy(self):
        assert "C" in vars(self.parameters)
        out = Matrix([0 for i in range(self.n_variables)])
        offset = self.level + 1
        b, h, alpha, beta, hinv = self.get_primitives()
        p = self.parameters
        tmp = 0
        for i in range(1 + self.level):
            for j in range(1 + self.level):
                tmp += alpha[i] * alpha[j]  + beta[i] * beta[j]
        sqrt = sympy.sqrt(tmp)
        for k in range(1 + self.level):
            for l in range(1 + self.level):
                out[1 + k] += (
                    -1.0 / (p.C**2 * self.basismatrices.M[k, k]) * alpha[l] * sqrt
                )
                if self.dimension == 2:
                    out[1 + k + offset] += (
                        -1.0 / (p.C**2 * self.basismatrices.M[k, k]) * beta[l] * sqrt
                    )
        return out
    
    def gravity(self):
        out = Matrix([0 for i in range(self.n_variables)])
        out[2] = -self.parameters.g * self.parameters.ex * self.variables[0]
        if self.dimension == 2:
            offset = self.level + 1
            out[2 + offset] = -self.parameters.g * self.parameters.ey * self.variables[0]
        return out
    
    def newtonian_turbulent_algebraic(self):
        assert "nu" in vars(self.parameters)
        assert "l_bl" in vars(self.parameters)
        assert "l_turb" in vars(self.parameters)
        assert "kappa" in vars(self.parameters)
        out = Matrix([0 for i in range(self.n_variables)])
        offset = self.level + 1
        
        b, h, a, b, hinv = self.get_primitives()
        
        p = self.parameters
        dU_dx = a[0] / (p.l_turb * h)
        abs_dU_dx = sympy.Piecewise((dU_dx, dU_dx >=0), (-dU_dx, True))
        for k in range(1 + self.level):
            out[1 + k] += (
                -(p.nu + p.kappa * sympy.sqrt(p.nu * abs_dU_dx) * p.l_bl * ( 1-p.l_bl)) * dU_dx * self.basismatrices.phib[k] * hinv
            )
            for i in range(1 + self.level):
                out[1 + k] += (
                    -p.nu * hinv
                    * a[i]
                    * self.basismatrices.D[i, k]
                )
                out[1 + k] += (
                    -p.kappa * sympy.sqrt(p.nu * abs_dU_dx) * hinv
                    * a[i]
                    * (self.basismatrices.Dxi[i, k] - self.basismatrices.Dxi2[i, k])
                )
        if self.dimension == 2:
            dV_dy = b[0] / (p.l_turb * h)
            abs_dV_dy = sympy.Piecewise((dV_dy, dV_dy >=0), (-dV_dy, True))
            for k in range(1 + self.level):
                out[1 + k + offset] += (
                    -(p.nu + p.kappa * sympy.sqrt(p.nu * abs_dV_dy) * p.l_bl * ( 1-p.l_bl)) * dV_dy * self.basismatrices.phib[k] *hinv
                )
                for i in range(1 + self.level):
                    out[1 + k + offset] += (
                        -p.nu
                        / h
                        * b[i]
                        * self.basismatrices.D[i, k]
                    )
                    out[1 + k + offset] += (
                    -p.kappa * sympy.sqrt(p.nu * abs_dV_dy) * hinv
                    * b[i]
                    * (self.basismatrices.Dxi[i, k] - self.basismatrices.Dxi2[i, k])
                )

        return out

@define(frozen=True, slots=True, kw_only=True)
class ShallowMomentsTopoNumerical(ShallowMomentsTopo):
    ref_model: Model = field(init=False)
    
    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        object.__setattr__(self, "ref_model", ShallowMomentsTopo(level=self.level, dimension=self.dimension, boundary_conditions=self.boundary_conditions))

    def flux(self):
        return [self.substitute_precomputed_denominator(f, self.variables[1], self.aux_variables.hinv) for f in self.ref_model.flux()]      
    
    def nonconservative_matrix(self):
        return [self.substitute_precomputed_denominator(f, self.variables[1], self.aux_variables.hinv) for f in self.ref_model.nonconservative_matrix()]  
    
    def quasilinear_matrix(self):
        return [self.substitute_precomputed_denominator(f, self.variables[1], self.aux_variables.hinv) for f in self.ref_model.quasilinear_matrix()]    
    
    def source(self):
        return self.substitute_precomputed_denominator(self.ref_model.source(), self.variables[1], self.aux_variables.hinv)
    
    def source_implicit(self):
        return self.substitute_precomputed_denominator(self.ref_model.source_implicit(), self.variables[1], self.aux_variables.hinv)
    
    def residual(self):
        return self.substitute_precomputed_denominator(self.ref_model.residual(), self.variables[1], self.aux_variables.hinv)
    
    def left_eigenvectors(self):
        return self.substitute_precomputed_denominator(self.ref_model.left_eigenvectors(), self.variables[1], self.aux_variables.hinv)
    
    def right_eigenvectors(self):
        return self.substitute_precomputed_denominator(self.ref_model.right_eigenvectors(), self.variables[1], self.aux_variables.hinv)

    def eigenvalues(self):
        h = self.variables[1]
        evs = self.substitute_precomputed_denominator(self.ref_model.eigenvalues(), self.variables[1], self.aux_variables.hinv)
        for i in range(self.n_variables):
            evs[i] = Piecewise((evs[i], h > 1e-8), (0, True))
        return evs



import numpy as np

"""
Dummy flux
"""


def Zero():
    def flux(Qi, Qj, Qauxi, Qauxj, param, normal, model_functions, mesh_props=None):
        Qout = np.zeros_like(Qi)
        return Qout, False

    return flux


"""
Lax-Friedrichs flux implementation
"""


def LF():
    def flux(Qi, Qj, Qauxi, Qauxj, param, normal, model_functions, mesh_props=None):
        assert mesh_props is not None
        dt_dx = mesh_props.dt_dx
        Qout = np.zeros_like(Qi)
        flux = model_functions.flux
        dim = normal.shape[0]
        num_eq = Qi.shape[0]
        for d in range(dim):
            Fi = flux[d](Qi, Qauxi, param)
            Fj = flux[d](Qj, Qauxj, param)
            Qout += 0.5 * (Fi + Fj) * normal[d]
        Qout -= 0.5 * dt_dx * (Qj - Qi)
        return Qout, False

    return flux


"""
Rusanov (local Lax-Friedrichs) flux implementation
"""


def LLF():
    def flux(Qi, Qj, Qauxi, Qauxj, param, normal, model_functions, mesh_props=None):
        EVi = model_functions.eigenvalues(Qi, Qauxi, param, normal)
        EVj = model_functions.eigenvalues(Qj, Qauxj, param, normal)
        assert not np.isnan(EVi).any()
        assert not np.isnan(EVj).any()
        smax = np.max(np.abs(np.vstack([EVi, EVj])))
        Qout = np.zeros_like(Qi)
        flux = model_functions.flux
        dim = normal.shape[0]
        num_eq = Qi.shape[0]
        for d in range(dim):
            Fi = flux[d](Qi, Qauxi, param)
            Fj = flux[d](Qj, Qauxj, param)
            Qout += 0.5 * (Fi + Fj) * normal[d]
        Qout -= 0.5 * smax * (Qj - Qi)
        return Qout, False

    return flux


"""
Rusanov (local Lax-Friedrichs) flux implementation
with topography fix (e.g. for model SWEtopo)
with WB fix for lake at rest
"""


def LLF_wb():
    def flux(Qi, Qj, Qauxi, Qauxj, param, normal, model_functions, mesh_props=None):
        IWB = np.eye(Qi.shape[0])
        IWB[-1, :] = 0.0
        IWB[0, -1] = 0.0
        EVi = np.zeros_like(Qi)
        EVj = np.zeros_like(Qj)
        EVi = model_functions.eigenvalues(Qi, Qauxi, param, normal)
        EVj = model_functions.eigenvalues(Qj, Qauxj, param, normal)
        assert not np.isnan(EVi).any()
        assert not np.isnan(EVj).any()
        smax = np.max(np.abs(np.vstack([EVi, EVj])))
        Qout = np.zeros_like(Qi)
        flux = model_functions.flux
        dim = normal.shape[0]
        num_eq = Qi.shape[0]
        Fi = np.zeros((num_eq))
        Fj = np.zeros((num_eq))
        for d in range(dim):
            flux[d](Qi, Qauxi, param, Fi)
            flux[d](Qj, Qauxj, param, Fj)
            Qout += 0.5 * (Fi + Fj) * normal[d]
        Qout -= 0.5 * smax * IWB @ (Qj - Qi)
        return Qout, False

    return flux



from sympy import Matrix, diff, exp, I, linear_eq_to_matrix, solve , Eq, zeros, simplify, nsimplify, latex,  symbols, Function, together, Symbol
from IPython.display import display, Latex


class ModelAnalyser():
    def __init__(self, model):
        self.model = model
        self.t = model.time
        x, y, z = model.position
        self.x = x
        self.y = y
        self.z = z
        self.equations = None
        self.plane_wave_symbols = []
        

    def get_equations(self):
        return self.equations

    def print_equations(self):          
        latex_lines = " \\\\\n".join([f"& {latex(eq)}" for eq in self.equations])
        latex_block = r"$$\begin{align*}" + "\n" + latex_lines + r"\end{align*}$$"
        display(Latex(latex_block))
            
    def get_time_space(self):
        x, y, z = self.model.position
        t = self.model.time
        return t, x, y, z
    
    def _get_omega_k(self):
        omega, kx, ky, kz = symbols('omega k_x k_y k_z')
        return omega, kx, ky, kz
    
    def _get_exponential(self):
        omega, kx, ky, kz = self._get_omega_k()
        t, x, y, z = self.get_time_space()
        exponential = exp(I * (kx * x + ky * y + kz * z - omega * t))
        return exponential
    
    def get_eps(self):
        eps = symbols('eps')
        return eps
    
    def create_functions_from_list(self, names):
        t, x, y, z = self.get_time_space()
        return [Function(name)(t, x, y, z) for name in names]
    
    def delete_equations(self, indices):
        self.equations = [self.equations[i] for i in range(len(self.equations)) if i not in indices]


    def solve_for_constraints(self, list_of_selected_equations, list_of_variables):
        equations = self.equations
        sol = solve([equations[i] for i in list_of_selected_equations], list_of_variables)
        equations = [eq.xreplace(sol).doit() for eq in equations]
        # delete used equations from equation system
        equations = [equations[i] for i in range(len(equations)) if i not in list_of_selected_equations]
        self.equations = equations
        return sol

    def insert_plane_wave_ansatz(self, functions_to_replace):
        exponential = self._get_exponential()
        f_bar_dict = {}
        for f in functions_to_replace:
            # Create the base name (e.g., 'f0')
            f_name = str(f.func)  # Get the function name (e.g., 'f0')
            
            # Create a new symbol representing \bar{f0}
            f_bar = Symbol(r'\bar{' + f_name + '}')
            f_bar_dict[f] =  f_bar * exponential
            self.plane_wave_symbols.append(f_bar)
        self.equations =  [eq.xreplace(f_bar_dict).doit() for eq in self.equations ]
        
    def solve_for_dispersion_relation(self):
        assert self.equations is not None, "No equations available to solve for dispersion relation."
        assert self.plane_wave_symbols, "No plane wave symbols available to solve for dispersion relation. Use insert_plane_wave_ansatz first."
        A, rhs = linear_eq_to_matrix(self.equations, self.plane_wave_symbols)
        omega, kx, ky, kz = self._get_omega_k()
        sol = solve(A.det(), omega)
        return sol

    def remove_exponential(self):
        exponential = self._get_exponential()
        equations = self.equations
        equations = [simplify(Eq(eq.lhs / exponential, eq.rhs / exponential)) for eq in equations]
        self.equations = equations

    def linearize_system(self, q, qaux, constraints=None):
        model = self.model
        t, x, y, z = self.get_time_space()
        dim = model.dimension
        X = [x, y, z]
        
        Q = Matrix(model.variables.get_list())
        Qaux = Matrix(model.aux_variables.get_list())

        
        substitutions = {Q[i]: q[i] for i in range(len(q))}
        substitutions.update({Qaux[i]: qaux[i] for i in range(len(qaux))})
        

        A = model.quasilinear_matrix()
        S = model.residual()
        if constraints is not None:
            C = constraints
        else:
            C = zeros(0, 1)
            

        Q = Q.xreplace(substitutions)
        for d in range(dim):
            A[d] = A[d].xreplace(substitutions)
        S = S.xreplace(substitutions)
        C = C.xreplace(substitutions)
        
        C = C.doit()
            
            
        gradQ = Matrix([diff(q[i], X[j]) for i in range(len(q)) for j in range(dim)]).reshape(len(q), dim)
            
        AgradQ = A[0] * gradQ[:, 0]
        for d in range(1, dim):
            AgradQ += A[d] * gradQ[:, d]


        expr = list(Matrix.vstack((diff(q, t) + AgradQ - S) , C))
        for i in range(len(expr)):
            expr[i] = nsimplify(expr[i], rational=True)
        expr = Matrix(expr)
        eps = self.get_eps()
        res = expr.copy()
        for i, e in enumerate(expr):
            collected = e
            collected = collected.series(eps, 0, 2).removeO()
            order_1_term = collected.coeff(eps, 1)
            res[i] = order_1_term
            
        for r in range(res.shape[0]):
            denom = together(res[r]).as_numer_denom()[1]
            res[r] *= denom
            res[r] = simplify(res[r])        
        
        linearized_system = [Eq((res[i]),0) for i in range(res.shape[0])]

        
        self.equations = linearized_system

